from datarobot.utils import encode_utf8_if_py2

from datarobot.models.api_object import APIObject
import trafaret as t


class MissingValuesReport(APIObject):
    """ Missing values report for model, contains list of reports per feature sorted by missing
    count in descending order.

    Notes
    -----

    ``Report per feature`` contains:

    * ``feature`` : feature name.
    * ``type`` : feature type -- 'Numeric' or 'Categorical'.
    * ``missing_count`` :  missing values count in training data.
    * ``missing_percentage`` : missing values percentage in training data.
    * ``tasks`` : list of information per each task, which was applied to feature.

    ``task information`` contains:

    * ``id`` : a number of task in the blueprint diagram.
    * ``name`` : task name.
    * ``descriptions`` : human readable aggregated information about how the task handles
      missing values.  The following descriptions may be present: what value is imputed for
      missing values, whether the feature being missing is treated as a feature by the task,
      whether missing values are treated as infrequent values,
      whether infrequent values are treated as missing values,
      and whether missing values are ignored.

    """
    _converter = t.Dict({
        t.Key('missing_values_report'):
            t.List(t.Dict({
                'feature': t.String,
                'type': t.String,
                'missing_count': t.Int,
                'missing_percentage': t.Float(),
                'tasks': t.Mapping(t.String, t.Dict({'name': t.String,
                                                     'descriptions': t.List(t.String)
                                                     }).ignore_extra('*')
                                   )
            }).ignore_extra('*'))
    }).ignore_extra('*')

    def __init__(self, missing_values_report):
        self._reports_per_feature = [MissingReportPerFeature(data) for data in
                                     missing_values_report]

    def __iter__(self):
        return iter(self._reports_per_feature)

    @classmethod
    def get(cls, project_id, model_id):
        """ Retrieve a missing report.

        Parameters
        ----------
        project_id : str
            The project's id.
        model_id : str
            The model's id.

        Returns
        -------
        MissingValuesReport
            The queried missing report.
        """
        url = 'projects/{}/models/{}/missingReport/'.format(project_id, model_id)
        return cls.from_location(url)


class TaskMissingReportInfo(object):
    """Convert dictionary into object, which contains task name, id (a number of task in the
    blueprint diagram) and descriptions (task specific missing report info).
    """

    def __init__(self, task_id, info):
        self.id = task_id
        self.name = info['name']
        self.descriptions = info['descriptions']

    def __repr__(self):
        return encode_utf8_if_py2(u'TaskMissingReportInfo(id={}, name={}, descriptions={})'
                                  .format(self.id, self.name, self.descriptions))

    def __eq__(self, other):
        return all([self.id == other.id,
                    self.name == other.name,
                    self.descriptions == other.descriptions])


class MissingReportPerFeature(object):
    """Convert dictionary into report per feature, which contains feature, type, missing count,
    percentage and tasks.
    """
    def __init__(self, report_per_feature_dict):
        self.feature = report_per_feature_dict['feature']
        self.type = report_per_feature_dict['type']
        self.missing_count = report_per_feature_dict['missing_count']
        self.missing_percentage = report_per_feature_dict['missing_percentage'] * 100
        self.tasks = [TaskMissingReportInfo(task_id, task_info) for task_id, task_info in
                      report_per_feature_dict['tasks'].items()]

    def __repr__(self):
        return encode_utf8_if_py2(u'MissingReportPerFeature(feature={},'
                                  u'type={}, miss_count={}, miss_percentage={}, tasks={}'
                                  .format(self.feature, self.type, self.missing_count,
                                          self.missing_percentage, self.tasks))

    def __eq__(self, other):
        return all([self.feature == other.feature,
                    self.type == other.type,
                    self.missing_count == other.missing_count,
                    self.missing_percentage == other.missing_percentage,
                    self.tasks == other.tasks])



from datetime import datetime

import dateutil
import pytest
import responses

from datarobot import BlenderModel


@responses.activate
@pytest.mark.usefixtures('client')
def test_blender_model_get_return_valid_data(blender_json, one_blender):
    url = 'https://host_name.com/projects/p-id/blenderModels/{}/'.format(one_blender.id)
    responses.add(responses.GET,
                  url,
                  body=blender_json,
                  status=200,
                  content_type='application/json')
    blender = BlenderModel.get('p-id', one_blender.id)
    assert isinstance(blender, BlenderModel)
    assert blender.id == one_blender.id
    assert blender.model_ids == one_blender.model_ids
    assert blender.blender_method == one_blender.blender_method


@pytest.fixture
def datetime_blender_data(datetime_model_data, blender_data):
    modified_data = dict(blender_data)
    modified_data.update(datetime_model_data)
    modified_data.pop('sample_pct')
    return modified_data


def test_datetime_blender_model(datetime_blender_data):
    mod = BlenderModel.from_data(datetime_blender_data)
    dt = mod.training_start_date
    assert dt.tzinfo == dateutil.tz.tzutc()
    assert isinstance(mod.training_end_date, datetime)
    assert mod.training_row_count is None
    assert mod.training_duration is None



import pytest
import responses
import json

from datarobot import Project
from datarobot.models.rating_table import RatingTable
from datarobot.errors import ClientError, InputNotUnderstoodError, JobAlreadyRequested
from datarobot.enums import JOB_TYPE


@pytest.fixture
def rating_table_url(project_collection_url, rating_table_backend_generated_json):
    d = json.loads(rating_table_backend_generated_json)
    return '{}{}/ratingTables/{}/'.format(project_collection_url, d['projectId'], d['id'])


@pytest.fixture
def downloaded_rating_table_contents():
    return b'''Matrix of word-grams occurrences tokenizer: scikit-learn based tokenizer
Matrix of word-grams occurrences binary: True
Matrix of word-grams occurrences sublinear_tf: False
Matrix of word-grams occurrences use_idf: False
Matrix of word-grams occurrences norm: L2
Intercept: -0.35839674349
Model precision: single
Loss distribution: Binomial Deviance
Link function: logit
Pairwise interactions found:
( number_emergency & number_inpatient ),0.0065962098320561146
( glyburide & number_emergency ),0.0016387530180316434

Feature Name,Feature Strength,Type,Transform1,Value1,Transform2,Value2,Weight,Coefficient
A1Cresult,0.0012348212995812102,CAT,One-hot,'>7',,,58.0,0.0082157774356443683
A1Cresult,0.0012348212995812102,CAT,One-hot,'>8',,,129.0,-0.012579536153936071
A1Cresult,0.0012348212995812102,CAT,One-hot,'Norm',,,59.0,0.032774696491885628
A1Cresult,0.0012348212995812102,CAT,One-hot,Missing value,,,1354.0,-0.00029931592903424328'''


@pytest.fixture
def downloaded_rating_table_file(temporary_file, downloaded_rating_table_contents):
    with open(temporary_file, 'wb') as out_f:
        out_f.write(downloaded_rating_table_contents)
    return temporary_file


@pytest.fixture
def completed_rating_table_validation_job(job_id, project_id, job_url):
    return {
        'status': 'COMPLETED',
        'url': job_url,
        'id': job_id,
        'jobType': JOB_TYPE.RATING_TABLE_VALIDATION,
        'isBlocked': False,
        'projectId': project_id
    }


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_table_backend_generated(rating_table_backend_generated_json,
                                            rating_table_backend_generated):
    backend_generated_id = rating_table_backend_generated.id
    url = 'https://host_name.com/projects/p-id/ratingTables/{}/'.format(backend_generated_id)

    responses.add(responses.GET,
                  url,
                  body=rating_table_backend_generated_json,
                  status=200,
                  content_type='application/json')

    table = RatingTable.get('p-id', backend_generated_id)
    assert isinstance(table, RatingTable)
    assert table.id == backend_generated_id
    assert not table.validation_job_id
    assert table.model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_table_not_modeled(rating_table_uploaded_but_not_modeled_json,
                                      rating_table_uploaded_but_not_modeled):
    not_modeled_id = rating_table_uploaded_but_not_modeled.id
    url = 'https://host_name.com/projects/p-id/ratingTables/{}/'.format(not_modeled_id)

    responses.add(responses.GET,
                  url,
                  body=rating_table_uploaded_but_not_modeled_json,
                  status=200,
                  content_type='application/json')
    table = RatingTable.get('p-id', not_modeled_id)
    assert isinstance(table, RatingTable)
    assert table.id == not_modeled_id
    assert table.validation_job_id
    assert not table.model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_table_modeled(rating_table_uploaded_and_modeled_json,
                                  rating_table_uploaded_and_modeled):
    modeled_id = rating_table_uploaded_and_modeled.id
    url = 'https://host_name.com/projects/p-id/ratingTables/{}/'.format(modeled_id)

    responses.add(responses.GET,
                  url,
                  body=rating_table_uploaded_and_modeled_json,
                  status=200,
                  content_type='application/json')
    table = RatingTable.get('p-id', modeled_id)
    assert isinstance(table, RatingTable)
    assert table.id == modeled_id
    assert table.validation_job_id
    assert table.model_id


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.usefixtures('known_warning')
def test_get_invalid_rating_table_is_fine_but_warns(invalid_rating_table_json,
                                                    invalid_rating_table):
    modeled_id = invalid_rating_table.id
    url = 'https://host_name.com/projects/p-id/ratingTables/{}/'.format(modeled_id)

    responses.add(responses.GET,
                  url,
                  body=invalid_rating_table_json,
                  status=200,
                  content_type='application/json')
    table = RatingTable.get('p-id', modeled_id)
    assert isinstance(table, RatingTable)
    assert table.id == modeled_id
    assert table.validation_job_id
    assert table.model_id


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.usefixtures('known_warning')
def test_get_invalid_rating_table_from_job_warns(project_url, rating_table_url,
                                                 invalid_rating_table_json,
                                                 downloaded_rating_table_file,
                                                 invalid_rating_table,
                                                 completed_rating_table_validation_job,
                                                 job_url):
    server_data = json.loads(invalid_rating_table_json)
    create_url = '{}ratingTables/'.format(project_url)
    responses.add(responses.POST,
                  create_url,
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET,
                  job_url,
                  json=completed_rating_table_validation_job,
                  status=303,
                  adding_headers={'Location': rating_table_url})
    responses.add(responses.GET,
                  rating_table_url,
                  body=invalid_rating_table_json,
                  status=200,
                  content_type='application/json')
    job = RatingTable.create(server_data['projectId'],
                             invalid_rating_table.model_id,
                             downloaded_rating_table_file)
    table = job.get_result_when_complete()
    assert isinstance(table, RatingTable)
    assert table.validation_job_id
    assert table.model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_with_invalid_rating_table_does_not_warn(project_url,
                                                      invalid_rating_table_json):
    server_data = json.loads(invalid_rating_table_json)
    list_url = '{}ratingTables/'.format(project_url)
    responses.add(responses.GET,
                  list_url,
                  body=json.dumps({'data': [server_data]}),
                  status=200,
                  content_type='application/json')
    rating_tables = Project(server_data['projectId']).get_rating_tables()
    table = rating_tables[0]
    assert isinstance(table, RatingTable)
    assert table.validation_job_id
    assert table.model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_table_not_found(rating_table_uploaded_and_modeled):
    modeled_id = rating_table_uploaded_and_modeled.id
    url = 'https://host_name.com/projects/p-id/ratingTables/{}/'.format(modeled_id)

    responses.add(responses.GET,
                  url,
                  body=json.dumps({"message": "This resource does not exist."}),
                  status=404,
                  content_type='application/json')
    with pytest.raises(ClientError):
        RatingTable.get('p-id', modeled_id)


@responses.activate
@pytest.mark.usefixtures('client')
def test_download_model_export(temporary_file, downloaded_rating_table_contents,
                               rating_table_uploaded_and_modeled):
    pid = rating_table_uploaded_and_modeled.project_id
    id = rating_table_uploaded_and_modeled.id
    url = 'https://host_name.com/projects/{}/ratingTables/{}/file'.format(pid, id)

    responses.add(responses.GET,
                  url,
                  body=downloaded_rating_table_contents)

    rating_table_uploaded_and_modeled.download(temporary_file)

    with open(temporary_file, 'rb') as in_f:
        saved_code = in_f.read()
    assert saved_code == downloaded_rating_table_contents


@responses.activate
@pytest.mark.usefixtures('client')
def test_create(project_url, base_job_server_data, rating_table_url,
                downloaded_rating_table_file, rating_table_backend_generated):
    url = '{}ratingTables/'.format(project_url)
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    responses.add(responses.POST,
                  url,
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET,
                  job_url,
                  json=base_job_server_data,
                  status=303,
                  adding_headers={'Location': rating_table_url})

    job = RatingTable.create(base_job_server_data['projectId'],
                             rating_table_backend_generated.model_id,
                             downloaded_rating_table_file)
    assert str(job.id) == base_job_server_data['id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_from_file(project_url, base_job_server_data, rating_table_url,
                          rating_table_backend_generated,
                          downloaded_rating_table_file):
    url = '{}ratingTables/'.format(project_url)
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    responses.add(responses.POST,
                  url,
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET,
                  job_url,
                  json=base_job_server_data,
                  status=303,
                  adding_headers={'Location': rating_table_url})

    job = RatingTable.create(base_job_server_data['projectId'],
                             rating_table_backend_generated.model_id,
                             downloaded_rating_table_file)
    assert str(job.id) == base_job_server_data['id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_invalid_sourcedata(base_job_server_data, rating_table_backend_generated):
    with pytest.raises(InputNotUnderstoodError):
        RatingTable.create(base_job_server_data['projectId'],
                           rating_table_backend_generated.model_id,
                           'Not real Rating Table content')


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_bad_parent_model(project_url, base_job_server_data,
                                 downloaded_rating_table_file):
    url = '{}ratingTables/'.format(project_url)
    responses.add(responses.POST,
                  url,
                  body=json.dumps({"message": "No existing rating table in the model"}),
                  status=400,
                  content_type='application/json')

    with pytest.raises(ClientError):
        RatingTable.create(base_job_server_data['projectId'],
                           '999999999999999999999999',
                           downloaded_rating_table_file)


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_model(rating_table_uploaded_but_not_modeled, base_job_server_data,
                      rating_table_model_url, project_url):
    url = '{}ratingTableModels/'.format(project_url)
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    responses.add(responses.POST,
                  url,
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET,
                  job_url,
                  json=base_job_server_data,
                  status=303,
                  adding_headers={'Location': rating_table_model_url})

    job = rating_table_uploaded_but_not_modeled.create_model()
    assert str(job.id) == base_job_server_data['id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_model_from_invalid_rating_table(project_url, base_job_server_data,
                                                rating_table_uploaded_but_not_modeled):
    url = '{}ratingTableModels/'.format(project_url)
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    responses.add(responses.POST,
                  url,
                  body=json.dumps({"message": "A prerequisite was not satisfied: Rating Table"
                                              " failed validation"}),
                  status=422,
                  content_type='application/json',
                  adding_headers={'Location': job_url})

    with pytest.raises(ClientError):
        rating_table_uploaded_but_not_modeled.create_model()


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_model_from_used_rating_table(project_url, base_job_server_data,
                                             rating_table_uploaded_and_modeled):
    url = '{}ratingTableModels/'.format(project_url)
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    responses.add(responses.POST,
                  url,
                  body=json.dumps({"message": "Rating Table already has a model",
                                   "errorName": "JobAlreadyAdded"}),
                  status=422,
                  content_type='application/json',
                  adding_headers={'Location': job_url})

    with pytest.raises(JobAlreadyRequested):
        rating_table_uploaded_and_modeled.create_model()


@responses.activate
@pytest.mark.usefixtures('client')
def test_rename_rating_table(rating_table_uploaded_but_not_modeled,
                             rating_table_renamed_json, renamed_rating_table):
    project_id = renamed_rating_table.project_id
    table_id = renamed_rating_table.id
    url = 'https://host_name.com/projects/{}/ratingTables/{}/'.format(project_id, table_id)
    responses.add(responses.PATCH,
                  url,
                  body=rating_table_renamed_json,
                  status=202,
                  content_type='application/json')
    table = rating_table_uploaded_but_not_modeled
    assert isinstance(table, RatingTable)
    assert table.rating_table_name != 'renamed'
    table.rename('renamed')
    assert table.rating_table_name == 'renamed'



# coding=utf-8
import datetime
import json
import re

import mock
import pandas as pd
import pytest
import responses
import six


from datarobot import (
    AdvancedOptions,
    AUTOPILOT_MODE,
    Blueprint,
    errors,
    Featurelist,
    Model,
    ModelJob,
    PredictJob,
    Project,
    QUEUE_STATUS,
    RatingTable,
    RatingTableModel,
    TARGET_TYPE,
    UserCV,
    UserTVH,
)
from datarobot.enums import PROJECT_STAGE, SCALEOUT_MODELING_MODE

from tests.test_helpers import URLParamsTestCase, fixture_file_path
from tests.utils import SDKTestcase, assert_raised_regex

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


# FIXTURES and HELPERS #############################################################################


def construct_dummy_aimed_project(
        pid='p-id',
        project_name='Untitled Project',
        target='SalePrice',
        autopilot_mode=0
):
    return {
        'id': pid,
        'projectName': project_name,
        'fileName': 'Untitled Project.csv',
        'stage': 'modeling',
        'autopilotMode': autopilot_mode,
        'created': '2015-05-21T16:02:02.573565',
        'target': target,
        'metric': 'Gamma Deviance',
        'partition': {
            'cvMethod': 'random',
            'validationType': 'CV',
            'validationPct': None,
            'holdoutPct': 20,
            'reps': 5,
            'holdoutLevel': None,
            'validationLevel': None,
            'trainingLevel': None,
            'partitionKeyCols': None,
            'userPartitionCol': None,
            'cvHoldoutLevel': None,
            'datetimeCol': None
        },
        'recommender': {
            'isRecommender': False,
            'recommenderItemId': None,
            'recommenderUserId': None
        },
        'advancedOptions': {
            'weights': None,
            'blueprintThreshold': None,
            'responseCap': False,
            'seed': None,
            'smartDownsampled': False,
            'majorityDownsamplingRate': None,
            'defaultMonotonicIncreasingFeaturelistId': None,
            'defaultMonotonicDecreasingFeaturelistId': None,
            'offset': None,
            'exposure': None
        },
        'positiveClass': None,
        'maxTrainPct': 64.0,
        'maxTrainRows': 128,
        'scaleoutMaxTrainPct': 64.0,
        'scaleoutMaxTrainRows': 128,
        'holdoutUnlocked': True,
        'targetType': 'Regression'
    }


AIMED_PROJECT_JSON = json.dumps(construct_dummy_aimed_project())


def prep_successful_project_creation_responses(project=None):
    """
    Setup the responses library to mock calls to the API necessary to
    create a project

    Parameters
    ----------
    project : dict
        Project JSON as a dict, as would be expected from construct_dummy_aimed_project_json
    """
    if project is None:
        project = construct_dummy_aimed_project()

    pid = project.get('id')
    responses.add(responses.POST,
                  'https://host_name.com/projects/',
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': 'https://host_name.com/status/status-id/'}
                  )
    responses.add(responses.GET,
                  'https://host_name.com/status/status-id/',
                  status=303,
                  body='',
                  content_type='application/json',
                  adding_headers={
                      'Location': 'https://host_name.com/projects/{}/'.format(pid),
                  }
                  )
    responses.add(responses.GET,
                  'https://host_name.com/projects/{}/'.format(pid),
                  status=200,
                  body=json.dumps(project),
                  content_type='application/json')


def prep_successful_aim_responses(project=None):
    """A helper to use with setting up test scenarios where the server is
    expected to successfully set the target.

    Parameters
    ----------
    project : dict
        Project JSON as a dict, as would be expected from construct_dummy_aimed_project_json
    """
    if project is None:
        project = construct_dummy_aimed_project()

    pid = project.get('id')
    project_url = 'https://host_name.com/projects/{}/'.format(pid)
    aim_url = project_url + 'aim/'

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': 'https://host_name.com/status/some-status/'},
                  content_type='application/json')

    responses.add(responses.GET,
                  'https://host_name.com/status/some-status/',
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=json.dumps(project),
                  status=200,
                  content_type='application/json')


# TESTS ############################################################################################


def test_future_proof(project_with_target_data):
    Project.from_data(dict(project_with_target_data, future='new'))


@pytest.mark.usefixtures('known_warning')
def test_project_from_dict_is_deprecated(project_without_target_data):
    Project(project_without_target_data)


class ProjectTestCase(SDKTestcase):
    def setUp(self):
        self.blueprint_api_resp = [
            {
                "projectId": "p-id",
                "processes": [
                    "Regularized Linear Model Preprocessing v5",
                    "Log Transformed Response"
                ],
                "id": "cbb4d6101dea1768ed79d75edd84c6c7",
                "modelType": "Auto-tuned K-Nearest Neighbors Regressor (Minkowski Distance)"
            },
            {
                "projectId": "p-id",
                "processes": [
                    "Regularized Linear Model Preprocessing v12",
                    "Log Transformed Response"
                ],
                "id": "e0708bd47f9fd21019a5ab7846e8364d",
                "modelType": "Auto-tuned Stochastic Gradient Descent Regression"
            },
        ]

    def test_instantiation_with_data(self):
        """
        Test instantiation Project(data)
        """
        data = {
            'id': 'project_id',
            'project_name': 'project_test_name',
            'mode': 2,
            'stage': 'stage',
            'target': 'test_target',
        }
        project_inst = Project.from_data(data)
        self.assertEqual(project_inst.id, data['id'])
        self.assertEqual(project_inst.project_name, data['project_name'])
        self.assertEqual(project_inst.mode, data['mode'])
        self.assertEqual(project_inst.stage, data['stage'])
        self.assertEqual(project_inst.target, data['target'])

        self.assertEqual(repr(project_inst), 'Project(project_test_name)')

    def test_print_project_nonascii_name(self):
        project = Project.from_data({
            'id': 'project-id',
            'project_name': u'\u3053\u3093\u306b\u3061\u306f',
        })
        print(project)

    def test_get_permalink(self):
        p = Project('pid')
        expected = 'https://host_name.com/projects/pid/models'
        self.assertEqual(expected, p.get_leaderboard_ui_permalink())

    @mock.patch('webbrowser.open')
    def test_open_leaderboard_browser(self, webbrowser_open):
        project = Project('p-id')
        project.open_leaderboard_browser()
        self.assertTrue(webbrowser_open.called)

    @responses.activate
    def test_create_project_async(self):
        prep_successful_project_creation_responses()

        new_project = Project.create(sourcedata='https://google.com')
        self.assertEqual(new_project.id, 'p-id')
        self.assertEqual(new_project.project_name, 'Untitled Project')

    @responses.activate
    def test_create_project_non_ascii_async(self):
        prep_successful_project_creation_responses()
        name = u'\xe3\x81\x82\xe3\x81\x82\xe3\x81\x82'
        Project.create("https://google.com", project_name=name)

    @responses.activate
    def test_get_project_metrics(self):
        """
        Test get project metrics
        """
        raw = """
        {"available_metrics":
            ["Gini Norm",
              "Weighted Gini Norm",
              "Weighted R Squared",
              "Weighted RMSLE",
              "Weighted MAPE",
              "Weighted Gamma Deviance",
              "Gamma Deviance",
              "RMSE",
              "Weighted MAD",
              "Tweedie Deviance",
              "MAD",
              "RMSLE",
              "Weighted Tweedie Deviance",
              "Weighted RMSE",
              "MAPE",
              "Weighted Poisson Deviance",
              "R Squared",
              "Poisson Deviance"],
         "feature_name": "SalePrice"}
        """
        expected_url = 'https://host_name.com/projects/p-id/features/metrics/'
        responses.add(responses.GET,
                      expected_url,
                      body=raw,
                      status=200,
                      content_type='application/json')
        get_project = Project('p-id').get_metrics('SalePrice')
        self.assertEqual(responses.calls[0].request.url,
                         expected_url + '?featureName=SalePrice')
        self.assertEqual(get_project["feature_name"], 'SalePrice')

    @responses.activate
    def test_get_project(self):
        """
        Test get project
        """
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/',
                      body=AIMED_PROJECT_JSON,
                      status=200,
                      content_type='application/json')
        get_project = Project.get('p-id')
        assert get_project.id == 'p-id'
        assert get_project.project_name == 'Untitled Project'
        assert get_project.target == 'SalePrice'
        assert get_project.target_type == 'Regression'
        assert get_project.stage == 'modeling'
        assert get_project.metric == 'Gamma Deviance'
        assert get_project.positive_class is None
        assert get_project.max_train_pct == 64.0
        assert get_project.max_train_rows == 128
        assert get_project.scaleout_max_train_pct == 64.0
        assert get_project.scaleout_max_train_rows == 128
        assert get_project.holdout_unlocked is True
        assert isinstance(get_project.partition, dict)
        assert isinstance(get_project.recommender, dict)
        assert isinstance(get_project.advanced_options, dict)
        assert 'weights' not in get_project.advanced_options
        assert 'offset' not in get_project.advanced_options
        assert 'exposure' not in get_project.advanced_options

        assert isinstance(get_project.created, datetime.datetime)

        assert get_project.partition['cv_method'] == 'random'

    @responses.activate
    def test_get_monotonicity_project(self):
        mono_config = {
            'onlyIncludeMonotonicBlueprints': True,
            'defaultMonotonicDecreasingFeaturelistId': '5ae04c0d962d7410683073cb',
            'defaultMonotonicIncreasingFeaturelistId': '5ae04c0f962d7410683073cc'
        }
        project_response = construct_dummy_aimed_project()
        project_response['advancedOptions'].update(**mono_config)
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/',
                      body=json.dumps(project_response),
                      status=200,
                      content_type='application/json')

        get_project = Project.get('p-id')

        assert get_project.advanced_options['default_monotonic_increasing_featurelist_id'] == \
            mono_config['defaultMonotonicIncreasingFeaturelistId']
        assert get_project.advanced_options['default_monotonic_decreasing_featurelist_id'] == \
            mono_config['defaultMonotonicDecreasingFeaturelistId']
        assert get_project.advanced_options['only_include_monotonic_blueprints'] == \
            mono_config['onlyIncludeMonotonicBlueprints']

    @responses.activate
    def test_get_project_weights_offset_exposure(self):
        """
        Test get project with a weights, offset and exposure columns
        """
        # given a project with a weights column
        pid = '582e4309100d2b13b939b223'
        weight_var = 'weight_variable'
        offset_var = ['offset_variable']
        exposure_var = 'exposure_variable'
        events_count_var = 'events_count'
        project_data = construct_dummy_aimed_project()
        project_data['advancedOptions']['weights'] = weight_var
        project_data['advancedOptions']['offset'] = offset_var
        project_data['advancedOptions']['exposure'] = exposure_var
        project_data['advancedOptions']['eventsCount'] = events_count_var
        # and a mocked API GET /projects/<pid>/ response
        responses.add(responses.GET,
                      'https://host_name.com/projects/{}/'.format(pid),
                      body=json.dumps(project_data),
                      status=200,
                      content_type='application/json')

        # when method Project.get is called
        project = Project.get(pid)

        # then the resulting object contains weights in advanced options
        assert project.advanced_options['weights'] == weight_var
        assert project.advanced_options['offset'] == offset_var
        assert project.advanced_options['exposure'] == exposure_var
        assert project.advanced_options['events_count'] == events_count_var

    @responses.activate
    def test_delete_project(self):
        """
        Test delete project
        """
        responses.add(responses.DELETE,
                      'https://host_name.com/projects/p-id/',
                      status=204)

        project = Project('p-id')
        del_result = project.delete()
        self.assertEqual(responses.calls[0].request.method, 'DELETE')
        self.assertIsNone(del_result)

    @patch('os.path.isfile')
    def test_non_ascii_filename(self, isfile_mock):
        isfile_mock.return_value = True
        bad_filename = u'\xc3\x98.csv'
        with pytest.raises(errors.IllegalFileName):
            Project.create(bad_filename)

    @responses.activate
    def test_create_non_ascii_filepath_is_okay(self):
        """As long as the file name is ascii, we're good"""
        prep_successful_project_creation_responses()
        path = fixture_file_path(u'日本/synthetic-100.csv')

        Project.create(path)
        # decoded to str implicitly
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/')
        request_message = responses.calls[0].request.body
        with open(path, 'rb') as fd:
            self.assertIn(fd.read(), request_message.to_string())

    @responses.activate
    def test_rename(self):
        responses.add(responses.PATCH,
                      'https://host_name.com/projects/p-id/',
                      body='',
                      status=200,
                      content_type='application/json')
        project = Project('p-id')
        project.rename('new name')
        assert responses.calls[0].request.method == 'PATCH'
        payload = json.loads(responses.calls[0].request.body)
        assert project.project_name == payload['projectName'] == 'new name'

    @responses.activate
    def test_unlock_holdout(self):
        responses.add(responses.PATCH,
                      'https://host_name.com/projects/p-id/',
                      body='',
                      status=200,
                      content_type='application/json')
        project = Project('p-id')
        upd_data = project.unlock_holdout()
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        payload = json.loads(responses.calls[0].request.body)
        self.assertTrue(payload['holdoutUnlocked'])
        self.assertTrue(upd_data)

    @responses.activate
    def test_set_worker_count(self):
        responses.add(responses.PATCH,
                      'https://host_name.com/projects/p-id/',
                      body='',
                      status=200,
                      content_type='application/json')
        project = Project('p-id')
        project.set_worker_count(20)
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        payload = json.loads(responses.calls[0].request.body)
        self.assertTrue(payload['workerCount'])

    @responses.activate
    def test_project_create_uses_read_timeout(self):
        """
        Check that the read_timeout parameter gets passed correctly down into
        the `requests` library

        The 400 is just because it's a simple way to short-circuit out of the project
        creation flow, which uses 3 total requests.
        """
        path = fixture_file_path('synthetic-100.csv')

        responses.add(responses.POST,
                      'https://host_name.com/projects/',
                      body='',
                      status=400
                      )
        with mock.patch.object(Project._client, 'request',
                               wraps=Project._client.request) as mock_request:
            try:
                Project.create(path, read_timeout=2)
            except errors.ClientError:
                pass

            timeout = mock_request.call_args[1]['timeout']
            assert timeout[0] == 6.05  # Connect timeout; specified in RESTClientObject
            assert timeout[1] == 2  # Read timeout; specified in the function call

    @responses.activate
    def test_by_upload_file_path(self):
        """
        Project.create(
            'synthetic-100.csv')
        """
        prep_successful_project_creation_responses()
        path = fixture_file_path('synthetic-100.csv')

        Project.create(path)
        # decoded to str implicitly
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/')
        request_message = responses.calls[0].request.body
        with open(path, 'rb') as fd:
            self.assertIn(fd.read(), request_message.to_string())

    def test_by_upload_file_path_fail(self):
        """
        Bad file name: Project.create('meh.csv')
        """
        path = fixture_file_path('does_not_exist.csv')
        with self.assertRaises(errors.InputNotUnderstoodError):
            Project.create(path)

    @responses.activate
    def test_by_upload_file_content(self):
        """
        Project.create(b'lalalala')
        """
        prep_successful_project_creation_responses()

        content_line = six.b('one,two\n' + '1,2' * 100)
        Project.create(content_line)
        # decoded to str implicitly
        request_message = responses.calls[0].request.body
        self.assertIn(content_line, request_message.to_string())

    def test_by_upload_content_encoding(self):
        """
        Bad content encoding Project.create(u'lalalala')
        """
        content_line = u'lalalala'
        with self.assertRaises(errors.InputNotUnderstoodError):
            Project.create(content_line)

    @responses.activate
    def test_by_upload_from_fd(self):
        """
        Project.create(
          sourcedata=open('synthetic-100.csv'))
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path('synthetic-100.csv')

        with open(path, 'rb') as fd:
            Project.create(sourcedata=fd)
            request_message = responses.calls[0].request.body

            with open(path, 'rb') as fd2:
                self.assertIn(fd2.read(), request_message.to_string())

        file_like = six.StringIO('thra\ntata\nrata')
        Project.create(
            sourcedata=file_like)

        # decoded to str implicitly
        request_message = responses.calls[3].request.body
        self.assertIn(six.b('thra\ntata\nrata'), request_message.to_string())

    @responses.activate
    def test_by_upload_file_seek(self):
        """
        Seek to EOF Project.create(
            open('synthetic-100.csv')
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path('synthetic-100.csv')
        with open(path, 'rb') as fd:
            fd.seek(20000000)
            Project.create(fd)
            # decoded to str implicitly

            request_message = responses.calls[0].request.body
            with open(path, 'rb') as fd2:
                self.assertIn(fd2.read(), request_message.to_string())

    @responses.activate
    def test_by_upload_file_closed(self):
        """
        Closed fd Project.create(
            open('synthetic-100.csv')
        """
        path = fixture_file_path('synthetic-100.csv')
        responses.add(responses.POST,
                      'https://host_name.com/projects/p-id/upload/',
                      body='',
                      status=200,
                      )
        fd = open(path)
        fd.close()
        with self.assertRaises(ValueError):
            Project.create(fd)

    @responses.activate
    def test_upload_by_file_url(self):
        """
        Project.create('http:/google.com/datarobot.csv')
        """
        prep_successful_project_creation_responses()

        link = 'http:/google.com/datarobot.csv'
        Project.create(sourcedata=link)
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            "url": link,
            "projectName": "Untitled Project",
        }
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/')

    @responses.activate
    def test_set_target(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        prep_successful_aim_responses()
        opts = AdvancedOptions(weights='WeightName',
                               offset=['OffsetName1', 'OffsetName2'],
                               events_count='events_count_col',
                               exposure='ExposureName')
        upd_project = Project('p-id').set_target(
            'SalePrice',
            metric='RMSE',
            advanced_options=opts)

        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'metric': 'RMSE',
            'weights': 'WeightName',
            'offset': ['OffsetName1', 'OffsetName2'],
            'exposure': 'ExposureName',
            'eventsCount': 'events_count_col',
            'smartDownsampled': False
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(upd_project, Project)
        advanced = upd_project.advanced_options
        assert advanced['response_cap'] is False
        assert advanced['smart_downsampled'] is False

    @responses.activate
    def test_set_target_price(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        prep_successful_aim_responses()
        upd_project = Project('p-id').set_target('SalePrice')

        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
        }
        self.assertEqual(upd_project.metric, 'Gamma Deviance')

    @responses.activate
    def test_set_target_quickrun_mode(self):
        prep_successful_aim_responses()
        Project('p-id').set_target('SalePrice', mode=AUTOPILOT_MODE.QUICK)

        request_body = json.loads(responses.calls[0].request.body)
        assert request_body['mode'] == AUTOPILOT_MODE.FULL_AUTO
        assert request_body['quickrun'] is True

    @responses.activate
    def test_set_target_with_valid_target_type_binary(self):
        valid_target_type = TARGET_TYPE.BINARY
        prep_successful_aim_responses()
        Project('p-id').set_target('SalePrice', target_type=valid_target_type)

        request_body = json.loads(responses.calls[0].request.body)
        assert request_body['targetType'] == valid_target_type

    @responses.activate
    def test_set_target_with_valid_target_type_regression(self):
        valid_target_type = TARGET_TYPE.REGRESSION
        prep_successful_aim_responses()
        Project('p-id').set_target('SalePrice', target_type=valid_target_type)

        request_body = json.loads(responses.calls[0].request.body)
        assert request_body['targetType'] == valid_target_type

    @responses.activate
    def test_set_target_with_valid_target_type_multiclass(self):
        valid_target_type = TARGET_TYPE.MULTICLASS
        prep_successful_aim_responses()
        Project('p-id').set_target('SalePrice', target_type=valid_target_type)

        request_body = json.loads(responses.calls[0].request.body)
        assert request_body['targetType'] == valid_target_type

    @responses.activate
    def test_set_target_with_invalid_target_type(self):
        invalid_target_type = 'blackbox_magic'
        prep_successful_aim_responses()

        with pytest.raises(TypeError) as exc_info:
            Project('p-id').set_target('SalePrice', target_type=invalid_target_type)
        assert_raised_regex(exc_info, 'is not a valid target_type')

        # Make sure that we did not actually make an invalid API call
        assert len(responses.calls) == 0

    @responses.activate
    def test_set_target_async_error(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        responses.add(responses.PATCH,
                      'https://host_name.com/projects/p-id/aim/',
                      body='',
                      status=202,
                      adding_headers={'Location': 'https://host_name.com/projects/p-id/'},
                      content_type='application/json')

        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/',
                      body=json.dumps({'status': 'ERROR'}),
                      status=200,
                      content_type='application/json')
        with self.assertRaises(errors.AsyncProcessUnsuccessfulError):
            Project('p-id').set_target(
                'SalePrice',
                metric='RMSE',
            )
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')

    def test_advanced_options_must_be_object(self):
        with self.assertRaises(TypeError):
            Project('p-id').set_target(
                'SalePrice',
                advanced_options={'garbage': 'in'}
            )

    @responses.activate
    def test_set_blueprint_threshold(self):
        """
        Set blueprint threshold
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(blueprint_threshold=2)
        upd_project = Project('p-id').set_target(
            'SalePrice',
            advanced_options=opts)

        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'blueprintThreshold': 2,
            'smartDownsampled': False
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_set_response_cap(self):
        """
        Set Response Cap
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(response_cap=0.7)
        upd_project = Project('p-id').set_target(
            'SalePrice',
            advanced_options=opts)

        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'responseCap': 0.7,
            'smartDownsampled': False
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_set_seed(self):
        """
        Set Response Cap
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(seed=22)
        upd_project = Project('p-id').set_target(
            'SalePrice',
            advanced_options=opts)

        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'seed': 22,
            'smartDownsampled': False
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_set_smart_sampled(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(smart_downsampled=True, majority_downsampling_rate=50.5)
        upd_project = Project('p-id').set_target('SalePrice', advanced_options=opts)

        request_message = responses.calls[0].request.body

        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'smartDownsampled': True,
            'majorityDownsamplingRate': 50.5,
        }
        assert isinstance(upd_project, Project)

    @responses.activate
    def test_set_target_advance_partition_method_cv(self):
        """
        Set project with advanced partition method
        """
        prep_successful_aim_responses()

        part_method = UserCV(user_partition_col='NumPartitions',
                             cv_holdout_level=1, seed=42)
        p = Project('p-id').set_target(
            'SalePrice',
            partitioning_method=part_method)
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'userPartitionCol': 'NumPartitions',
            'cvHoldoutLevel': 1,
            'seed': 42,
            'validationType': 'CV',
            'cvMethod': 'user',
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(p, Project)

    @responses.activate
    def test_set_target_advance_partition_method_tvh(self):
        """
        Set project with advanced partition method
        """
        prep_successful_aim_responses()
        part_method = UserTVH(user_partition_col='NumPartitions',
                              validation_level=1, training_level=2,
                              holdout_level=3,
                              seed=42)
        p = Project('p-id').set_target(
            'SalePrice',
            partitioning_method=part_method)
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'userPartitionCol': 'NumPartitions',
            'holdoutLevel': 3,
            'seed': 42,
            'validationLevel': 1,
            'trainingLevel': 2,
            'validationType': 'TVH',
            'cvMethod': 'user'
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(p, Project)

    @responses.activate
    def test_set_target_specify_positive_class(self):
        """
        Set project with advanced partition method
        """
        prep_successful_aim_responses()

        p = Project('p-id').set_target(
            'Forks',
            positive_class='None or Unspecified')
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'Forks',
            'positiveClass': 'None or Unspecified',
            'mode': AUTOPILOT_MODE.FULL_AUTO
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(p, Project)

    @responses.activate
    @pytest.mark.usefixtures('known_warning')
    def test_set_target_quickrun_param(self):
        """
        Set project with quickrun option
        """
        prep_successful_aim_responses()

        p = Project('p-id').set_target(
            'Forks', mode=AUTOPILOT_MODE.MANUAL,
            quickrun=True)
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'Forks',
            'quickrun': True,
            'mode': AUTOPILOT_MODE.FULL_AUTO
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(p, Project)

    def test_pass_advance_part_wrong(self):
        with self.assertRaises(TypeError):
            Project('p-id').set_target(
                'SalePrice',
                partitioning_method={'CV': 'TVH'})

    @responses.activate
    def test_pause_autopilot(self):
        """
        Project('p-id').pause_autopilot()
        """
        responses.add(responses.POST,
                      'https://host_name.com/projects/p-id/autopilot/',
                      body='',
                      status=202,
                      content_type='application/json')
        self.assertTrue(Project('p-id').pause_autopilot())
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {'command': 'stop'}

    @responses.activate
    def test_unpause_autopilot(self):
        """
        Project('p-id').unpause_autopilot()
        """
        responses.add(responses.POST,
                      'https://host_name.com/projects/p-id/autopilot/',
                      body='',
                      status=202,
                      content_type='application/json')
        self.assertTrue(Project('p-id').unpause_autopilot())
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {'command': 'start'}

    @responses.activate
    def test_start_autopilot(self):
        """
        Project('p-id').start_autopilot('featurelist-id')
        """
        responses.add(responses.POST,
                      'https://host_name.com/projects/p-id/autopilots/',
                      body='',
                      status=201,
                      content_type='application/json')
        Project('p-id').start_autopilot('featurelist-id')
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'featurelistId': 'featurelist-id',
            'mode': AUTOPILOT_MODE.FULL_AUTO
        }

    @responses.activate
    def test_get_featurelists(self):
        """project.get_featurelists()

        """
        some_featurelists = [
            {
                'id': 'f-id-1',
                'projectId': 'p-id',
                'name': 'Raw Features',
                'features': ['One Fish', 'Two Fish', 'Red Fish', 'Blue Fish'],
            },
            {
                'id': 'f-id-2',
                'projectId': 'p-id',
                'name': 'Informative Features',
                'features': ['One Fish', 'Red Fish', 'Blue Fish'],
            },
            {
                'id': 'f-id-3',
                'projectId': 'p-id',
                'name': 'Custom Features',
                'features': ['One Fish', 'Blue Fish'],
            },
        ]

        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/featurelists/',
                      body=json.dumps(some_featurelists),
                      status=200,
                      content_type='application/json')
        flists = Project('p-id').get_featurelists()
        for flist in flists:
            self.assertIsInstance(flist, Featurelist)

    def test_create_featurelist_duplicate_features(self):
        project = Project('p-id')
        with pytest.raises(errors.DuplicateFeaturesError) as exc_info:
            project.create_featurelist('test', ['feature', 'feature'])
        assert_raised_regex(exc_info, 'Can\'t create featurelist with duplicate features')

    @responses.activate
    def test_start_project(self):
        """
        Project.start("test_name",
                      "file.csv",
                      "a_target",
                      worker_count=4,
                      metric="a_metric")
        """
        path = fixture_file_path('synthetic-100.csv')
        prep_successful_project_creation_responses()

        responses.add(responses.PATCH,
                      'https://host_name.com/projects/p-id/',
                      body='',
                      status=200,
                      content_type='application/json')

        prep_successful_aim_responses()

        proj = Project.start(project_name="test_name",
                             sourcedata=path,
                             target="a_target",
                             worker_count=4,
                             metric="a_metric")
        self.assert_project_start_call_order()
        self.assertIsInstance(proj, Project)

    @responses.activate
    def test_start_project_from_dataframe(self):
        """
        Project.start("test_name",
                      "file.csv",
                      "a_target",
                      worker_count=4,
                      metric="a_metric")
        """
        prep_successful_project_creation_responses()

        responses.add(responses.PATCH,
                      'https://host_name.com/projects/p-id/',
                      body='',
                      status=200,
                      content_type='application/json')

        prep_successful_aim_responses()

        dataframe = pd.DataFrame({
            'a_target': range(100),
            'predictor': range(100, 200),
        })
        proj = Project.start(dataframe,
                             "a_target",
                             "test_name",
                             worker_count=4,
                             metric="a_metric")
        self.assert_project_start_call_order()
        self.assertIsInstance(proj, Project)

    @responses.activate
    def test_start_project_in_manual_mode(self):
        """
        Project.start("test_name",
                      "file.csv",
                      "a_target",
                      worker_count=4,
                      metric="a_metric",
                      autopilot_on=False)
        """
        path = fixture_file_path('synthetic-100.csv')
        prep_successful_project_creation_responses()
        responses.add(responses.PATCH,
                      'https://host_name.com/projects/p-id/',
                      body='',
                      status=200,
                      content_type='application/json')
        prep_successful_aim_responses()

        proj = Project.start(path,
                             "test_name",
                             "a_target",
                             worker_count=4,
                             metric="a_metric",
                             autopilot_on=False)
        self.assert_project_start_call_order()
        self.assertIsInstance(proj, Project)

        req_body = json.loads(responses.calls[4].request.body)
        self.assertEqual(req_body['mode'], AUTOPILOT_MODE.MANUAL)

    @responses.activate
    def test_set_scaleout_mode(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        prep_successful_aim_responses()
        opts = AdvancedOptions(scaleout_modeling_mode=SCALEOUT_MODELING_MODE.AUTOPILOT)
        upd_project = Project('p-id').set_target(
            'SalePrice',
            metric='RMSE',
            advanced_options=opts)

        request_message = json.loads(responses.calls[0].request.body)
        self.assertEqual(request_message, {
            'target': 'SalePrice',
            'mode': AUTOPILOT_MODE.FULL_AUTO,
            'metric': 'RMSE',
            'smartDownsampled': False,
            'scaleoutModelingMode': SCALEOUT_MODELING_MODE.AUTOPILOT
        })
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(upd_project, Project)

    def assert_project_start_call_order(self):
        """Run this assertion to assert that the expected calls for project.start
        happened, and in the correct order

        This is a terribly brittle test. If we can come up with something
        better, let's go with that.
        """
        # Create Project
        self.assertEqual(responses.calls[0].request.url,
                         'https://host_name.com/projects/')
        self.assertEqual(responses.calls[0].request.method, 'POST')

        # Get Project Creation Status
        self.assertEqual(responses.calls[1].request.url,
                         'https://host_name.com/status/status-id/')
        self.assertEqual(responses.calls[1].request.method, 'GET')

        # Get Created Project
        self.assertEqual(responses.calls[2].request.url,
                         'https://host_name.com/projects/p-id/')
        self.assertEqual(responses.calls[2].request.method, 'GET')

        # Set Worker Count
        self.assertEqual(responses.calls[3].request.url,
                         'https://host_name.com/projects/p-id/')
        self.assertEqual(responses.calls[3].request.method, 'PATCH')

        # Set Target
        self.assertEqual(responses.calls[4].request.url,
                         'https://host_name.com/projects/p-id/aim/')
        self.assertEqual(responses.calls[4].request.method, 'PATCH')

        # Get Status of Finalizing Project
        self.assertEqual(responses.calls[5].request.url,
                         'https://host_name.com/status/some-status/')
        self.assertEqual(responses.calls[5].request.method, 'GET')

        # Get Finalized Project
        self.assertEqual(responses.calls[6].request.url,
                         'https://host_name.com/projects/p-id/')
        self.assertEqual(responses.calls[6].request.method, 'GET')

    @responses.activate
    def test_set_target_featurelist(self):
        """
        proj.set_target('test_target', featurelist_id=...)
        """
        prep_successful_aim_responses()
        Project('p-id').set_target('a_target', featurelist_id='f-id-3')
        body = json.loads(responses.calls[0].request.body)
        assert body['featurelistId'] == 'f-id-3'

    @responses.activate
    def test_set_target_monotonicity(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(
            monotonic_increasing_featurelist_id='fi-id',
            monotonic_decreasing_featurelist_id='fd-id',
            only_include_monotonic_blueprints=True
        )
        p = Project('p-id').set_target(
            'Forks',
            positive_class='None or Unspecified',
            advanced_options=opts
        )
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            'target': 'Forks',
            'mode': 'auto',
            'smartDownsampled': False,
            'monotonicIncreasingFeaturelistId': 'fi-id',
            'monotonicDecreasingFeaturelistId': 'fd-id',
            'onlyIncludeMonotonicBlueprints': True,
            'positiveClass': 'None or Unspecified'
        }
        self.assertEqual(responses.calls[0].request.method, 'PATCH')
        self.assertEqual(responses.calls[1].request.method, 'GET')
        self.assertIsInstance(p, Project)

    @responses.activate
    def test_attach_file_with_link_goes_to_url(self):
        """
        Project.create('https://google.com/file.csv')
        """
        prep_successful_project_creation_responses()

        link = 'http:/google.com/datarobot.csv'
        Project.create(link)
        request_message = responses.calls[0].request.body
        assert json.loads(request_message) == {
            "url": link,
            "projectName": "Untitled Project",
        }
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/')

    @responses.activate
    def test_attach_file_with_file_path(self):
        """
        Project.create('synthetic-100.csv')
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path('synthetic-100.csv')
        Project.create(path)

        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/')
        request_message = responses.calls[0].request.body
        with open(path, 'rb') as fd:
            self.assertIn(fd.read(), request_message.read())

    @responses.activate
    def test_attach_file_with_non_csv_file_path(self):
        """
        Project.create('dataset.xlsx')
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path('onehundredrows.xlsx')
        Project.create(path)
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/')
        request_message = responses.calls[0].request.body.read()
        with open(path, 'rb') as fd:
            self.assertIn(fd.read(), request_message)
        fname = re.search(b'filename="(.*)"',
                          request_message).group(1)
        self.assertEqual(fname, b'onehundredrows.xlsx')

    @responses.activate
    def test_get_status_underscorizes(self):
        body = json.dumps({
            'autopilotDone': True,
            'stage': 'modeling',
            'stageDescription': 'Ready for modeling',
        })
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/status/',
                      status=200,
                      body=body,
                      content_type='application/json'
                      )
        status = Project('p-id').get_status()
        self.assertEqual(status['stage'], 'modeling')
        self.assertTrue(status['autopilot_done'])

    @responses.activate
    def test_get_blueprints(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/blueprints/',
            status=200,
            body=json.dumps(self.blueprint_api_resp),
            content_type='application/json'
        )
        menu = Project('p-id').get_blueprints()
        for item in menu:
            self.assertIsInstance(item, Blueprint)
            self.assertEqual(item.project_id, 'p-id')
        bluepr1 = menu[0]
        bluepr2 = menu[1]
        self.assertIsInstance(bluepr1.processes, list)
        self.assertEqual(bluepr1.processes[0],
                         'Regularized Linear Model Preprocessing v5')
        self.assertEqual(bluepr1.processes[1], 'Log Transformed Response')
        self.assertEqual(bluepr1.model_type,
                         'Auto-tuned K-Nearest Neighbors Regressor (Minkowski Distance)')
        self.assertIsInstance(bluepr2.processes, list)
        self.assertEqual(bluepr2.processes[0],
                         'Regularized Linear Model Preprocessing v12')
        self.assertEqual(bluepr2.processes[1],
                         'Log Transformed Response')
        self.assertEqual(bluepr2.model_type,
                         'Auto-tuned Stochastic Gradient Descent Regression')

    @responses.activate
    def test_get_monotonicity_blueprints(self):
        mono_config = {
            'supportsMonotonicConstraints': True,
            'monotonicDecreasingFeaturelistId': '5ae04c0d962d7410683073cb',
            'monotonicIncreasingFeaturelistId': '5ae04c0f962d7410683073cc'
        }
        blueprint_resp = [dict(**b) for b in self.blueprint_api_resp]
        [b.update(mono_config) for b in blueprint_resp]
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/blueprints/',
            status=200,
            body=json.dumps(blueprint_resp),
            content_type='application/json'
        )
        blueprints = Project('p-id').get_blueprints()
        for blueprint in blueprints:
            assert blueprint.monotonic_increasing_featurelist_id == \
                mono_config['monotonicIncreasingFeaturelistId']
            assert blueprint.monotonic_decreasing_featurelist_id == \
                mono_config['monotonicDecreasingFeaturelistId']
            assert blueprint.supports_monotonic_constraints == \
                mono_config['supportsMonotonicConstraints']


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_featurelist(featurelist_server_data, project_id):
    project = Project(project_id)
    name = featurelist_server_data['name']
    features = featurelist_server_data['features']

    responses.add(responses.POST,
                  'https://host_name.com/projects/{}/featurelists/'.format(project_id),
                  json=featurelist_server_data,
                  status=201,
                  content_type='application/json',
                  adding_headers={
                      'Location': 'https://host_name.com/projects/{}/featurelists/{}/'.format(
                          project_id, featurelist_server_data['id']
                      )
                  })
    new_flist = project.create_featurelist(name, features)
    assert new_flist.name == name
    assert new_flist.features == features
    assert new_flist.project_id == project_id
    assert new_flist.num_models == featurelist_server_data['numModels']
    assert new_flist.is_user_created == featurelist_server_data['isUserCreated']
    assert new_flist.description == featurelist_server_data['description']
    assert isinstance(new_flist.created, datetime.datetime)


class TestProjectJobListing(SDKTestcase):
    job1_queued_dict = {
        'status': 'queue',
        'processes': [
            'Majority Class Classifier'
        ],
        'projectId': '556902e8100d2b3728d47551',
        'samplePct': 32.0,
        'modelType': 'Majority Class Classifier',
        'featurelistId': '556902eb100d2b37d1130771',
        'blueprintId': '89e08076a908e859c07af49bd4aa6a0f',
        'id': '10',
        'isBlocked': False,
        'modelId': '556902ef100d2b37da13077c'
    }
    job2_queued_dict = {
        'status': 'queue',
        'processes': [
            'One-Hot Encoding',
            'Missing Values Imputed',
            'RuleFit Classifier'
        ],
        'projectId': '556902e8100d2b3728d47551',
        'samplePct': 32.0,
        'modelType': 'RuleFit Classifier',
        'featurelistId': '556902eb100d2b37d1130771',
        'blueprintId': 'a8959bc1d46f07fb3dc14db7c1e3fc99',
        'id': '11',
        'isBlocked': False,
        'modelId': '556902ef100d2b37da13077d'
    }
    job3_queued_dict = {
        'status': 'queue',
        'processes': [
            'One-Hot Encoding',
            'Missing Values Imputed',
            'RuleFit Classifier'
        ],
        'projectId': '556902e8100d2b3728d47551',
        'samplePct': 64.0,
        'modelType': 'RuleFit Classifier',
        'featurelistId': '556902eb100d2b37d1130771',
        'blueprintId': 'a8959bc1d46f07fb3dc14db7c1e3fc99',
        'id': '17',
        'isBlocked': False,
        'modelId': '556902ef100d2b37da13077d'
    }
    job1_inprogress_dict = dict(job1_queued_dict, status='inprogress')
    job2_inprogress_dict = dict(job2_queued_dict, status='inprogress')

    predict_job_queued_dict = {
        'status': 'queue',
        'projectId': '56b62892ccf94e7e939c89c8',
        'message': '',
        'id': '27',
        'isBlocked': False,
        'modelId': '56b628b7ccf94e0444bb8152'
    }

    predict_job_errored_dict = {
        'status': 'error',
        'projectId': '56b62892ccf94e7e939c89c8',
        'message': '',
        'id': '23',
        'isBlocked': False,
        'modelId': '56b628b7ccf94e0444bb8152'
    }

    all_jobs_response = {
        'count': 4,
        'next': None,
        'jobs': [
            {
                'status': 'inprogress',
                'url': 'https://host_name.com/projects/p-id/modelJobs/1/',
                'id': '1',
                'jobType': 'model',
                'isBlocked': False,
                'projectId': 'p-id'
            },
            {
                'status': 'inprogress',
                'url': 'https://host_name.com/projects/p-id/modelJobs/2/',
                'id': '2',
                'jobType': 'model',
                'isBlocked': False,
                'projectId': 'p-id'
            },
            {
                'status': 'queue',
                'url': 'https://host_name.com/projects/p-id/predictJobs/3/',
                'id': '3',
                'jobType': 'predict',
                'isBlocked': False,
                'projectId': 'p-id'
            },
            {
                'status': 'queue',
                'url': 'https://host_name.com/projects/p-id/modelJobs/4/',
                'id': '4',
                'jobType': 'model',
                'isBlocked': False,
                'projectId': 'p-id'
            }
        ],
        'previous': None
    }

    all_error_jobs_response = {
        'count': 1,
        'next': None,
        'jobs': [
            {
                'status': 'error',
                'url': 'https://host_name.com/projects/p-id/predictJobs/3/',
                'id': '3',
                'jobType': 'predict',
                'isBlocked': False,
                'projectId': 'p-id'
            }
        ],
        'previous': None
    }

    @responses.activate
    def test_get_all_jobs(self):
        body = json.dumps(self.all_jobs_response)
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/jobs/',
                      status=200,
                      body=body,
                      content_type='application/json'
                      )
        jobs = Project('p-id').get_all_jobs()
        self.assertEqual(len(jobs), 4)

    @responses.activate
    def test_get_all_errored_jobs(self):
        body = json.dumps(self.all_error_jobs_response)
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/jobs/',
                      status=200,
                      body=body,
                      content_type='application/json'
                      )
        jobs = Project('p-id').get_all_jobs(status=QUEUE_STATUS.ERROR)
        self.assertEqual(responses.calls[0].request.url,
                         'https://host_name.com/projects/p-id/jobs/?status=error')
        self.assertEqual(len(jobs), 1)

    @responses.activate
    def test_get_model_jobs(self):
        job_dict = self.job1_queued_dict
        body = json.dumps([job_dict])
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/modelJobs/',
                      status=200,
                      body=body,
                      content_type='application/json'
                      )
        jobs = Project('p-id').get_model_jobs(status=QUEUE_STATUS.QUEUE)
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/p-id/modelJobs/?status=queue')
        self.assertEqual(len(jobs), 1)
        self.assertDictEqual(jobs[0].__dict__, ModelJob(job_dict).__dict__)

    @responses.activate
    def test_get_model_jobs_requests_all_by_default(self):
        job_dict = self.job1_queued_dict
        body = json.dumps([job_dict])
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/modelJobs/',
                      status=200,
                      body=body,
                      content_type='application/json'
                      )
        jobs = Project('p-id').get_model_jobs()
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/p-id/modelJobs/')
        self.assertEqual(len(jobs), 1)
        self.assertDictEqual(jobs[0].__dict__, ModelJob(job_dict).__dict__)

    @responses.activate
    def test_get_predict_jobs(self):
        body = json.dumps([self.predict_job_errored_dict])
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/predictJobs/',
                      status=200,
                      body=body,
                      content_type='application/json')
        jobs = Project('p-id').get_predict_jobs(status=QUEUE_STATUS.ERROR)
        self.assertEqual(responses.calls[0].request.url,
                         'https://host_name.com/projects/p-id/predictJobs/?status=error')
        self.assertEqual(len(jobs), 1)
        result_job = jobs[0]
        expected_job = PredictJob(self.predict_job_errored_dict)
        self.assertEqual(expected_job.id, result_job.id)
        self.assertEqual(expected_job.project, result_job.project)

    @responses.activate
    def test_get_predict_jobs_default(self):
        body = json.dumps([self.predict_job_queued_dict])
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/predictJobs/',
                      status=200,
                      body=body,
                      content_type='application/json')
        jobs = Project('p-id').get_predict_jobs()
        self.assertEqual(responses.calls[0].request.url,
                         'https://host_name.com/projects/p-id/predictJobs/')
        self.assertEqual(len(jobs), 1)
        result_job = jobs[0]
        expected_job = PredictJob(self.predict_job_queued_dict)
        self.assertEqual(expected_job.id, result_job.id)
        self.assertEqual(expected_job.project, result_job.project)

    @patch('datarobot.Project.get_model_jobs')
    def test_job_status_counter(self, mock_get_jobs):
        project = Project('p-id')
        jobs0 = []
        jobs1 = [ModelJob(self.job1_queued_dict)]
        jobs2 = [ModelJob(self.job1_inprogress_dict), ModelJob(self.job2_queued_dict)]
        jobs3 = [ModelJob(self.job2_inprogress_dict)]
        mock_get_jobs.side_effect = [jobs0, jobs1, jobs2, jobs3]
        assert project._get_job_status_counts() == (0, 0)
        assert project._get_job_status_counts() == (0, 1)
        assert project._get_job_status_counts() == (1, 1)
        assert project._get_job_status_counts() == (1, 0)

    def test_wait_for_autopilot(self):
        def make_status(autopilot_done):
            return {'autopilot_done': autopilot_done}

        project = Project('p-id')
        with mock.patch.object(Project, '_autopilot_status_check') as mock_autopilot_check:
            mock_autopilot_check.side_effect = [make_status(False)] * 3 + \
                                               [make_status(True)] + \
                                               [make_status(False)]

            project.wait_for_autopilot(check_interval=.5, verbosity=0)
            assert mock_autopilot_check.call_count == 4

    def test_wait_for_autopilot_timeout(self):
        project = Project('p-id')
        with mock.patch.object(Project, '_autopilot_status_check') as mock_autopilot_check:
            mock_autopilot_check.return_value = {'autopilot_done': False}
            self.assertRaises(errors.AsyncTimeoutError,
                              project.wait_for_autopilot,
                              check_interval=10, verbosity=0, timeout=.2)

    def test_wait_for_autopilot_target_not_set_is_error(self):
        project = Project('p-id')
        with mock.patch.object(Project, 'get_status') as mock_get_status:
            mock_get_status.return_value = {
                'autopilot_done': False,
                'stage': PROJECT_STAGE.AIM,
            }
            with pytest.raises(RuntimeError) as exc_info:
                project.wait_for_autopilot(check_interval=10, verbosity=0, timeout=.2)
            assert_raised_regex(exc_info, 'target has not been set')

    def test_wait_for_autopilot_mode_incorrect_is_error(self):
        project = Project('p-id')
        with mock.patch.multiple(
                Project,
                get_status=lambda _self: {
                    'autopilot_done': False,
                    'stage': PROJECT_STAGE.MODELING,
                },
                _server_data=lambda _self, _: construct_dummy_aimed_project(autopilot_mode=2)):
            with pytest.raises(RuntimeError) as exc_info:
                project.wait_for_autopilot(check_interval=10, verbosity=0, timeout=.2)

            assert_raised_regex(exc_info, 'mode is not full auto')


class TestProjectList(SDKTestcase):
    def setUp(self):
        self.raw_return = """
        [
            {
            "project_name": "Api project",
            "_id": "54c627fa100d2b2c7002a489"
            },
            {
            "_id": "54c78125100d2b2fe3b296b6",
            "project_name": "Untitled project"
            }
        ]
        """

    @responses.activate
    def test_list_projects(self):
        """
        Test list projects
        """
        responses.add(responses.GET,
                      'https://host_name.com/projects/',
                      body=self.raw_return,
                      status=200,
                      content_type='application/json')
        project_lists = Project.list()
        assert isinstance(project_lists, list)
        assert project_lists[0].id == '54c627fa100d2b2c7002a489'
        assert project_lists[0].project_name == 'Api project'
        assert project_lists[1].id == '54c78125100d2b2fe3b296b6'
        assert project_lists[1].project_name == 'Untitled project'

    @responses.activate
    def test_with_manual_search_params(self):
        responses.add(responses.GET,
                      'https://host_name.com/projects/?projectName=x',
                      body=self.raw_return,
                      status=200,
                      content_type='application/json',
                      match_querystring=True)
        Project('p-id').list(search_params={'project_name': 'x'})
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/?projectName=x')

    @responses.activate
    def test_with_bad_search_params(self):
        with self.assertRaises(TypeError):
            Project('p-id').list(search_params=12)


class TestGetModels(SDKTestcase, URLParamsTestCase):
    def setUp(self):
        super(TestGetModels, self).setUp()
        self.raw_data = """
        [
    {
    "featurelistId": "556cdfbd100d2b10048c7941",
    "processes": ["One", "Two", "Three"],
    "featurelistName": "Informative Features",
    "projectId": "p-id",
    "samplePct": 64,
    "modelType": "AVG Blender",
    "metrics": {
        "AUC": {
            "holdout": 0.76603,
            "validation": 0.64141,
            "crossValidation": 0.7625240000000001
        },
        "Rate@Top5%": {
            "holdout": 1,
            "validation": 0.5,
            "crossValidation": 0.9
        },
        "Rate@TopTenth%": {
            "holdout": 1,
            "validation": 1,
            "crossValidation": 1
        },
        "RMSE": {
            "holdout": 0.42054,
            "validation": 0.44396,
            "crossValidation": 0.40162000000000003
        },
        "LogLoss": {
            "holdout": 0.53707,
            "validation": 0.58051,
            "crossValidation": 0.5054160000000001
        },
        "FVE Binomial": {
            "holdout": 0.17154,
            "validation": 0.03641,
            "crossValidation": 0.17637399999999998
        },
        "Gini Norm": {
            "holdout": 0.53206,
            "validation": 0.28282,
            "crossValidation": 0.525048
        },
        "Rate@Top10%": {
            "holdout": 1,
            "validation": 0.25,
            "crossValidation": 0.7
        }
    },
    "blueprintId": "a4fd9d17a8ca62ee00590dd704dae6a8",
    "id": "556ce973100d2b6e51ca9657"
},
    {
    "featurelistId": "556cdfbd100d2b10048c7941",
    "processes": ["One", "Two", "Three"],
    "featurelistName": "Informative Features",
    "projectId": "p-id",
    "samplePct": 64,
    "modelType": "AVG Blender",
    "metrics": {
        "AUC": {
            "holdout": 0.76603,
            "validation": 0.64141,
            "crossValidation": 0.7625240000000001
        },
        "Rate@Top5%": {
            "holdout": 1,
            "validation": 0.5,
            "crossValidation": 0.9
        },
        "Rate@TopTenth%": {
            "holdout": 1,
            "validation": 1,
            "crossValidation": 1
        },
        "RMSE": {
            "holdout": 0.42054,
            "validation": 0.44396,
            "crossValidation": 0.40162000000000003
        },
        "LogLoss": {
            "holdout": 0.53707,
            "validation": 0.58051,
            "crossValidation": 0.5054160000000001
        },
        "FVE Binomial": {
            "holdout": 0.17154,
            "validation": 0.03641,
            "crossValidation": 0.17637399999999998
        },
        "Gini Norm": {
            "holdout": 0.53206,
            "validation": 0.28282,
            "crossValidation": 0.525048
        },
        "Rate@Top10%": {
            "holdout": 1,
            "validation": 0.25,
            "crossValidation": 0.7
        }
    },
    "blueprintId": "a4fd9d17a8ca62ee00590dd704dae6a8",
    "id": "556ce973100d2b6e51ca9658"
}
]
        """

    @responses.activate
    def test_get_models_ordered_by_metric_by_default(self):
        """
        Project('p-id').get_models()
        """
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/models/?orderBy=-metric',
                      body=self.raw_data,
                      status=200,
                      content_type='application/json',
                      match_querystring=True)
        leaderboard = Project('p-id').get_models()
        self.assertNotEqual(0, len(leaderboard))
        for item in leaderboard:
            self.assertIsInstance(item, Model)
        self.assertEqual(leaderboard[0].id, '556ce973100d2b6e51ca9657')
        self.assertEqual(leaderboard[1].id, '556ce973100d2b6e51ca9658')
        self.assertEqual(leaderboard[0].project_id, 'p-id')
        self.assertEqual(leaderboard[1].project_id, 'p-id')

    @responses.activate
    def test_get_models_ordered_by_specific_field(self):
        # ordering
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/models/?orderBy=samplePct',
                      body=self.raw_data,
                      status=200,
                      content_type='application/json',
                      match_querystring=True)
        Project('p-id').get_models(order_by='samplePct')
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/p-id/models/?orderBy=samplePct')

    @responses.activate
    def test_get_models_orderd_by_two_fields(self):
        # ordering by two fields
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/models/?orderBy=samplePct%2C-metric',
            body=self.raw_data,
            status=200,
            content_type='application/json',
            match_querystring=True)
        Project('p-id').get_models(
            order_by=['sample_pct', '-metric'])
        self.assertEqual(
            responses.calls[0].request.url,
            'https://host_name.com/projects/p-id/models/?orderBy=samplePct%2C-metric')

    @responses.activate
    def test_get_models_two_order_fields_and_filtering(self):
        # ordering by two fields plus filtering
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/models/'
            '?withMetric=RMSE&orderBy=metric%2C-samplePct',
            body=self.raw_data,
            status=200,
            content_type='application/json',
            match_querystring=True)
        Project('p-id').get_models(
            order_by=['metric', '-sample_pct'], with_metric='RMSE')

    @responses.activate
    def test_get_monotonicity_models(self):
        mono_config = {
            'supportsMonotonicConstraints': True,
            'monotonicIncreasingFeaturelistId': '5ae04c0d962d7410683073cb',
            'monotonicDecreasingFeaturelistId': '5ae04c0f962d7410683073cc'
        }
        models_resp = json.loads(self.raw_data)
        [m.update(mono_config) for m in models_resp]
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/models/'
            '?withMetric=RMSE&orderBy=-metric',
            body=json.dumps(models_resp),
            status=200,
            content_type='application/json',
            match_querystring=True)

        models = Project('p-id').get_models(with_metric='RMSE')

        for model in models:
            assert model.supports_monotonic_constraints == \
                mono_config['supportsMonotonicConstraints']
            assert model.monotonic_increasing_featurelist_id == \
                mono_config['monotonicIncreasingFeaturelistId']
            assert model.monotonic_decreasing_featurelist_id == \
                mono_config['monotonicDecreasingFeaturelistId']

    def test_order_by_with_unexpected_param_fails(self):
        with pytest.raises(ValueError) as exc_info:
            Project('p-id').get_models(order_by='someThing')
        assert_raised_regex(exc_info, 'Provided order_by attribute')

    def test_order_by_with_bad_param_fails(self):
        with pytest.raises(TypeError) as exc_info:
            Project('p-id').get_models(order_by=True)
        assert_raised_regex(exc_info, 'Provided order_by attribute')

    def test_with_bad_search_param_fails(self):
        with self.assertRaises(TypeError):
            Project('p-id').get_models(search_params=True)

    def _canonize_order_by_handles_none(self):
        proj = Project('p-id')
        self.assertIsNone(proj._canonize_order_by(None))

    @responses.activate
    def test_get_models_search_starred(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/p-id/models/?orderBy=-metric&isStarred=false',
            body=self.raw_data,
            status=200,
            content_type='application/json',
            match_querystring=True)
        Project('p-id').get_models(search_params={'is_starred': False})

        requested_url = responses.calls[0].request.url
        address, params = requested_url.split('?')
        assert address == 'https://host_name.com/projects/p-id/models/'
        params_list = params.split('&')
        assert len(params_list) == 2
        assert 'orderBy=-metric' in params_list
        assert 'isStarred=false' in params_list


@responses.activate
@pytest.mark.usefixtures('client')
def retrieve_smart_sampled_project(project_url, project_id, smart_sampled_project_server_data):
    responses.add(responses.GET, project_url,
                  body=json.dumps(smart_sampled_project_server_data),
                  status=200, content_type='application/json')
    smart_sampled_project = Project.get(project_id)
    advanced = smart_sampled_project.advanced_options
    assert advanced['smart_downsampled'] is True
    rate_key = 'majority_downsampling_rate'
    assert advanced[rate_key] == smart_sampled_project_server_data[rate_key]


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_tables(rating_table_backend_generated_json,
                           rating_table_uploaded_but_not_modeled_json,
                           rating_table_uploaded_and_modeled_json):

    sample_rating_tables = json.dumps(
        {'data': [json.loads(rating_table_backend_generated_json),
                  json.loads(rating_table_uploaded_but_not_modeled_json),
                  json.loads(rating_table_uploaded_and_modeled_json)]})

    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/ratingTables/',
                  body=sample_rating_tables,
                  status=200,
                  content_type='application/json')
    rating_tables = Project('p-id').get_rating_tables()

    assert len(rating_tables) == 3
    for rt in rating_tables:
        assert isinstance(rt, RatingTable)
        assert rt.id


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_tables_no_tables():

    sample_rating_tables = json.dumps({'data': []})

    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/ratingTables/',
                  body=sample_rating_tables,
                  status=200,
                  content_type='application/json')
    rating_tables = Project('p-id').get_rating_tables()

    assert len(rating_tables) == 0


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_table_models(rating_table_model_server_data):
    sample_rating_table_model = json.dumps([rating_table_model_server_data,
                                            rating_table_model_server_data])

    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/ratingTableModels/',
                  body=sample_rating_table_model,
                  status=200,
                  content_type='application/json')
    rating_table_models = Project('p-id').get_rating_table_models()

    assert len(rating_table_models) == 2
    for rtm in rating_table_models:
        assert isinstance(rtm, RatingTableModel)
        assert rtm.id
        assert rtm.rating_table_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_rating_table_models_no_models():
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/ratingTableModels/',
                  body='[]',
                  status=200,
                  content_type='application/json')
    rating_table_models = Project('p-id').get_rating_table_models()

    assert len(rating_table_models) == 0



import json

import pytest
import responses
from datarobot import Model

from datarobot.models.lift_chart import LiftChart


@pytest.fixture
def lift_chart_validation_data():
    return {
        'source': 'validation',
        'bins': [
            {'binWeight': 2.0, 'actual': 19.0, 'predicted': 18.701785714285712},
            {'binWeight': 1.0, 'actual': 10.5, 'predicted': 9.716000000000001},
            {'binWeight': 1.0, 'actual': 7.5, 'predicted': 9.73913043478261},
            {'binWeight': 2.0, 'actual': 17.5, 'predicted': 19.561085972850677},
            {'binWeight': 1.0, 'actual': 12.1, 'predicted': 10.4},
            {'binWeight': 1.0, 'actual': 13.8, 'predicted': 10.549999999999997},
            {'binWeight': 2.0, 'actual': 28.5, 'predicted': 23.451428571428572},
            {'binWeight': 1.0, 'actual': 17.8, 'predicted': 13.057142857142855},
            {'binWeight': 1.0, 'actual': 14.4, 'predicted': 13.452631578947367},
            {'binWeight': 2.0, 'actual': 27.4, 'predicted': 29.022926829268297},
            {'binWeight': 1.0, 'actual': 23.2, 'predicted': 15.760377358490558},
            {'binWeight': 1.0, 'actual': 17.2, 'predicted': 15.811538461538465},
            {'binWeight': 2.0, 'actual': 29.799999999999997, 'predicted': 31.967142857142857},
            {'binWeight': 1.0, 'actual': 21.7, 'predicted': 16.33478260869565},
            {'binWeight': 1.0, 'actual': 13.1, 'predicted': 17.022222222222226},
            {'binWeight': 2.0, 'actual': 33.9, 'predicted': 34.43989743589744},
            {'binWeight': 1.0, 'actual': 20.5, 'predicted': 17.86857142857143},
            {'binWeight': 1.0, 'actual': 15.6, 'predicted': 18.06},
            {'binWeight': 2.0, 'actual': 37.400000000000006, 'predicted': 36.72619047619048},
            {'binWeight': 1.0, 'actual': 20.4, 'predicted': 18.79},
            {'binWeight': 1.0, 'actual': 19.6, 'predicted': 19.18068181818182},
            {'binWeight': 2.0, 'actual': 39.0, 'predicted': 39.42089047903004},
            {'binWeight': 1.0, 'actual': 21.9, 'predicted': 20.055714285714277},
            {'binWeight': 1.0, 'actual': 21.8, 'predicted': 20.089583333333334},
            {'binWeight': 2.0, 'actual': 45.7, 'predicted': 40.20238095238094},
            {'binWeight': 1.0, 'actual': 21.8, 'predicted': 20.117894736842103},
            {'binWeight': 1.0, 'actual': 16.2, 'predicted': 20.24895833333333},
            {'binWeight': 2.0, 'actual': 38.6, 'predicted': 40.828123261764375},
            {'binWeight': 1.0, 'actual': 20.0, 'predicted': 20.56131386861314},
            {'binWeight': 1.0, 'actual': 24.3, 'predicted': 20.604385964912293},
            {'binWeight': 2.0, 'actual': 41.5, 'predicted': 41.352748538011696},
            {'binWeight': 1.0, 'actual': 19.8, 'predicted': 20.710743801652903},
            {'binWeight': 1.0, 'actual': 21.1, 'predicted': 20.7235294117647},
            {'binWeight': 2.0, 'actual': 32.6, 'predicted': 41.49125},
            {'binWeight': 1.0, 'actual': 18.6, 'predicted': 20.78782608695652},
            {'binWeight': 1.0, 'actual': 21.0, 'predicted': 20.794690265486736},
            {'binWeight': 2.0, 'actual': 37.3, 'predicted': 42.73298850574714},
            {'binWeight': 1.0, 'actual': 22.3, 'predicted': 21.538356164383554},
            {'binWeight': 1.0, 'actual': 23.0, 'predicted': 21.79859154929577},
            {'binWeight': 2.0, 'actual': 45.5, 'predicted': 44.31785714285715},
            {'binWeight': 1.0, 'actual': 22.6, 'predicted': 22.854999999999997},
            {'binWeight': 1.0, 'actual': 22.4, 'predicted': 23.418749999999996},
            {'binWeight': 2.0, 'actual': 47.0, 'predicted': 47.75747028862478},
            {'binWeight': 1.0, 'actual': 24.7, 'predicted': 24.20416666666667},
            {'binWeight': 1.0, 'actual': 16.5, 'predicted': 24.233333333333334},
            {'binWeight': 2.0, 'actual': 44.0, 'predicted': 49.43414634146342},
            {'binWeight': 1.0, 'actual': 23.1, 'predicted': 24.863829787234042},
            {'binWeight': 1.0, 'actual': 26.5, 'predicted': 25.13},
            {'binWeight': 2.0, 'actual': 50.9, 'predicted': 51.14829424307035},
            {'binWeight': 1.0, 'actual': 30.7, 'predicted': 27.419999999999998},
            {'binWeight': 1.0, 'actual': 24.8, 'predicted': 28.247999999999998},
            {'binWeight': 2.0, 'actual': 51.900000000000006, 'predicted': 58.22245989304814},
            {'binWeight': 1.0, 'actual': 24.0, 'predicted': 29.263636363636362},
            {'binWeight': 1.0, 'actual': 33.2, 'predicted': 31.84444444444444},
            {'binWeight': 2.0, 'actual': 65.6, 'predicted': 64.13333333333334},
            {'binWeight': 1.0, 'actual': 34.9, 'predicted': 33.515},
            {'binWeight': 1.0, 'actual': 29.0, 'predicted': 34.068749999999994},
            {'binWeight': 2.0, 'actual': 68.8, 'predicted': 69.67843137254901},
            {'binWeight': 1.0, 'actual': 42.8, 'predicted': 43.05},
            {'binWeight': 2.0, 'actual': 88.7, 'predicted': 95.475}]
    }


@pytest.fixture
def expected_bin_data(lift_chart_validation_data):
    expected_bins = [dict(bin_data) for bin_data in lift_chart_validation_data['bins']]
    for expected in expected_bins:
        weight = expected.pop('binWeight')
        expected['bin_weight'] = weight
    return expected_bins


def test_instantiation(lift_chart_validation_data, expected_bin_data):
    lc = LiftChart.from_server_data(lift_chart_validation_data)

    assert lc.source == lift_chart_validation_data['source']
    assert lc.bins == expected_bin_data


def test_future_proof(lift_chart_validation_data):
    data_with_future_keys = dict(lift_chart_validation_data, new_key='some future lift data')
    data_with_future_keys['bins'][0]['new_key'] = 'some future bin data'
    LiftChart.from_server_data(data_with_future_keys)


@pytest.fixture
def lift_chart_validation_data_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/liftChart/validation/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def lift_chart_list_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/liftChart/'.format(
        project_id,
        model_id
    )


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_validation_lift_chart(lift_chart_validation_data,
                                   expected_bin_data,
                                   lift_chart_validation_data_url,
                                   project_id, model_id):
    responses.add(
        responses.GET,
        lift_chart_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(lift_chart_validation_data)
    )
    model = Model(id=model_id, project_id=project_id)
    lc = model.get_lift_chart('validation')

    assert lc.source == lift_chart_validation_data['source']
    assert lc.bins == expected_bin_data


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_all_lift_charts(lift_chart_validation_data,
                             expected_bin_data,
                             lift_chart_list_url,
                             project_id, model_id):
    responses.add(
        responses.GET,
        lift_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [lift_chart_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id)
    lc_list = model.get_all_lift_charts()

    assert len(lc_list) == 1
    assert lc_list[0].source == lift_chart_validation_data['source']
    assert lc_list[0].bins == expected_bin_data



from .partitioning_methods import *  # noqa


class AdvancedOptions(object):
    """
    Used when setting the target of a project to set advanced options of modeling process.

    Parameters
    ----------
    weights : string, optional
        The name of a column indicating the weight of each row
    response_cap : float in [0.5, 1), optional
        Quantile of the response distribution to use for response capping.
    blueprint_threshold : int, optional
        Number of hours models are permitted to run before being excluded from later autopilot
        stages
        Minimum 1
    seed : int
        a seed to use for randomization
    smart_downsampled : bool
        whether to use smart downsampling to throw away excess rows of the majority class.  Only
        applicable to classification and zero-boosted regression projects.
    majority_downsampling_rate : float
        the percentage between 0 and 100 of the majority rows that should be kept.  Specify only if
        using smart downsampling.  May not cause the majority class to become smaller than the
        minority class.
    offset : list of str, optional
        (New in version v2.6) the list of the names of the columns containing the offset
        of each row
    exposure : string, optional
        (New in version v2.6) the name of a column containing the exposure of each row
    accuracy_optimized_mb : bool, optional
        (New in version v2.6) Include additional, longer-running models that will be run by the
        autopilot and available to run manually.
    scaleout_modeling_mode : string, optional
        (New in version v2.8) Specifies the behavior of Scaleout models for the project.
        This is one of ``datarobot.enums.SCALEOUT_MODELING_MODE``.
        If ``datarobot.enums.SCALEOUT_MODELING_MODE.DISABLED``, no
        models will run during autopilot or show in the list of available blueprints.
        Scaleout models must be disabled for some partitioning settings including projects
        using datetime partitioning or projects using offset or exposure columns.
        If ``datarobot.enums.SCALEOUT_MODELING_MODE.REPOSITORY_ONLY``,
        scaleout models will be in the list of available blueprints
        but not run during autopilot.
        If ``datarobot.enums.SCALEOUT_MODELING_MODE.AUTOPILOT``,
        scaleout models will run during autopilot and be in the list of
        available blueprints.
        Scaleout models are only supported in the Hadoop enviroment with
        the corresponding user permission set.
    events_count : string, optional
        (New in version v2.8) the name of a column specifying events count.
    monotonic_increasing_featurelist_id : string, optional
        (new in version 2.11) the id of the featurelist that defines the set of features
        with a monotonically increasing relationship to the target. If None,
        no such constraints are enforced. When specified, this will set a default for the project
        that can be overriden at model submission time if desired.
    monotonic_decreasing_featurelist_id : string, optional
        (new in version 2.11) the id of the featurelist that defines the set of features
        with a monotonically decreasing relationship to the target. If None,
        no such constraints are enforced. When specified, this will set a default for the project
        that can be overriden at model submission time if desired.
    only_include_monotonic_blueprints : bool, optional
        (new in version 2.11) when true, only blueprints that support enforcing
        monotonic constraints will be available in the project or selected for the autopilot.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        advanced_options = dr.AdvancedOptions(
            weights='weights_column',
            offset=['offset_column'],
            exposure='exposure_column',
            response_cap=0.7,
            blueprint_threshold=2,
            smart_downsampled=True, majority_downsampling_rate=75.0)

    """
    def __init__(self, weights=None, response_cap=None, blueprint_threshold=None,
                 seed=None, smart_downsampled=False, majority_downsampling_rate=None,
                 offset=None, exposure=None, accuracy_optimized_mb=None,
                 scaleout_modeling_mode=None, events_count=None,
                 monotonic_increasing_featurelist_id=None, monotonic_decreasing_featurelist_id=None,
                 only_include_monotonic_blueprints=None):
        self.weights = weights
        self.response_cap = response_cap
        self.blueprint_threshold = blueprint_threshold
        self.seed = seed
        self.smart_downsampled = smart_downsampled
        self.majority_downsampling_rate = majority_downsampling_rate
        self.offset = offset
        self.exposure = exposure
        self.accuracy_optimized_mb = accuracy_optimized_mb
        self.scaleout_modeling_mode = scaleout_modeling_mode
        self.events_count = events_count
        self.monotonic_increasing_featurelist_id = monotonic_increasing_featurelist_id
        self.monotonic_decreasing_featurelist_id = monotonic_decreasing_featurelist_id
        self.only_include_monotonic_blueprints = only_include_monotonic_blueprints

    def collect_payload(self):

        payload = dict(
            weights=self.weights,
            response_cap=self.response_cap,
            blueprint_threshold=self.blueprint_threshold,
            seed=self.seed,
            smart_downsampled=self.smart_downsampled,
            majority_downsampling_rate=self.majority_downsampling_rate,
            offset=self.offset,
            exposure=self.exposure,
            accuracy_optimized_mb=self.accuracy_optimized_mb,
            scaleout_modeling_mode=self.scaleout_modeling_mode,
            events_count=self.events_count,
            monotonic_increasing_featurelist_id=self.monotonic_increasing_featurelist_id,
            monotonic_decreasing_featurelist_id=self.monotonic_decreasing_featurelist_id,
            only_include_monotonic_blueprints=self.only_include_monotonic_blueprints
        )
        return payload



import pandas
import pytest
import responses

from mock import patch

from datarobot import ReasonCodesInitialization, ReasonCodes
from datarobot.models.reason_codes import ReasonCodesPage, ReasonCodesRow


@pytest.fixture
def rci_url(project_collection_url, rci_data):
    return '{}{}/models/{}/reasonCodesInitialization/'.format(project_collection_url,
                                                              rci_data['project_id'],
                                                              rci_data['model_id'])


def test_reason_codes_initialization_future_proof(rci_data):
    ReasonCodesInitialization.from_data(dict(rci_data, future='new'))


def test_reason_codes_future_proof(rc_data):
    ReasonCodes.from_data(dict(rc_data, future='new'))


def test_reason_codes_page_future_proof(rcp_data):
    with patch('datarobot.models.reason_codes.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        ReasonCodesPage.from_data(dict(rcp_data, future='new'))


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_initialization_get_returns_valid_data(rci_url, rci_json, one_rci):
    responses.add(responses.GET,
                  rci_url,
                  body=rci_json,
                  status=200,
                  content_type='application/json')
    initialization = ReasonCodesInitialization.get(one_rci.project_id, one_rci.model_id)
    assert isinstance(initialization, ReasonCodesInitialization)
    assert initialization.project_id == one_rci.project_id
    assert initialization.model_id == one_rci.model_id
    assert initialization.reason_codes_sample


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_initialization_create(rci_data, rci_url, project_url, base_job_server_data):
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    responses.add(responses.POST, rci_url,
                  body='', status=202, content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET, job_url,
                  json=base_job_server_data, status=303, adding_headers={'Location': rci_url})

    rci_job = ReasonCodesInitialization.create(project_id=rci_data['project_id'],
                                               model_id=rci_data['model_id'])
    assert str(rci_job.id) == base_job_server_data['id']


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_initialization_delete(rci_url, rci_data):
    responses.add(responses.DELETE, rci_url, status=200)

    rci = ReasonCodesInitialization.from_data(rci_data)
    rci.delete()


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_get_returns_valid_data(project_id, project_url, rc_json, one_rc):
    url = '{}reasonCodesRecords/{}/'.format(project_url, one_rc.id)
    responses.add(responses.GET,
                  url,
                  body=rc_json,
                  status=200,
                  content_type='application/json')
    record = ReasonCodes.get(project_id, one_rc.id)
    assert isinstance(record, ReasonCodes)
    assert record.id == one_rc.id
    assert record.model_id == one_rc.model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_reason_codes_page_get_returns_valid_data(project_id, project_url, rcp_json, one_rcp):
    url = '{}reasonCodes/{}/'.format(project_url, one_rcp.id)
    responses.add(responses.GET,
                  url,
                  body=rcp_json,
                  status=200,
                  content_type='application/json')
    with patch('datarobot.models.reason_codes.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        record = ReasonCodesPage.get(project_id, one_rcp.id)
    assert isinstance(record, ReasonCodesPage)
    assert record.id == one_rcp.id
    assert record.data == one_rcp.data
    assert record.adjustment_method == 'N/A'


@responses.activate
@pytest.mark.usefixtures('client')
def test_reason_codes_page_get_returns_valid_data_adjusted(
        project_id, project_url, rcp_data_with_adjusted_predictions, one_rcp):
    url = '{}reasonCodes/{}/'.format(project_url, one_rcp.id)
    responses.add(responses.GET,
                  url,
                  json=rcp_data_with_adjusted_predictions,
                  status=200,
                  content_type='application/json')
    with patch('datarobot.models.reason_codes.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        record = ReasonCodesPage.get(project_id, one_rcp.id)
    assert isinstance(record, ReasonCodesPage)
    assert record.id == one_rcp.id
    assert record.data == one_rcp.data
    assert record.adjustment_method == 'exposureNormalized'
    for rc_row in record.data:
        assert rc_row['adjusted_prediction'] == rc_row['prediction']
        assert rc_row['adjusted_prediction_values'] == rc_row['prediction_values']


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_create(project_id, project_url, project_collection_url,
                             base_job_server_data, rcp_data):
    rcp_base_url = '{}{}/reasonCodes/'.format(project_collection_url, project_id)
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    rcp_url = '{}/{}/'.format(rcp_base_url, rcp_data['id'])

    responses.add(responses.POST, rcp_base_url,
                  body='', status=202, content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET, job_url,
                  json=base_job_server_data, status=303, adding_headers={'Location': rcp_url})

    rcp_job = ReasonCodes.create(project_id=project_id,
                                 model_id='578e59a41ced2e5a9eb18960',
                                 dataset_id='581c6256100d2b60586980d9')
    assert str(rcp_job.id) == base_job_server_data['id']


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_list(project_id, project_url, rc_list_json):
    rc_base_url = '{}reasonCodesRecords/'.format(project_url)
    responses.add(responses.GET, rc_base_url + '?modelId=578e59a41ced2e5a9eb18965',
                  status=200, body=rc_list_json, content_type='application/json',
                  match_querystring=True)
    rc_list = ReasonCodes.list(project_id, model_id='578e59a41ced2e5a9eb18965')
    assert len(rc_list) == 3
    for rc in rc_list:
        assert isinstance(rc, ReasonCodes)


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_delete(project_collection_url, project_id, rc_data):
    rc_url = '{}{}/reasonCodesRecords/{}/'.format(project_collection_url,
                                                  project_id,
                                                  rc_data['id'])
    responses.add(responses.DELETE, rc_url, status=200)

    rc = ReasonCodes.from_data(rc_data)
    rc.delete()


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_iteration(project_id, project_url, one_rcp):
    url = '{}reasonCodes/{}/'.format(project_url, one_rcp.id)
    url_page1 = url + '?offset=0&excludeAdjustedPredictions=true'
    url_page2 = url + '?limit=2&offset=2&excludeAdjustedPredictions=true'
    url_page3 = url + '?limit=2&offset=4&excludeAdjustedPredictions=true'
    resp_data_1 = {'id': one_rcp.id,
                   'count': 5,
                   'previous': None,
                   'next': url_page2,
                   'data': one_rcp.data[:2],
                   'reason_codes_record_location': one_rcp.reason_codes_record_location}
    resp_data_2 = dict(resp_data_1, previous=url_page1, next=url_page3, data=one_rcp.data[2:4])
    resp_data_3 = dict(resp_data_1, previous=url_page2, next=None, data=one_rcp.data[4:5])
    responses.add(responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True)
    responses.add(responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True)
    responses.add(responses.GET, url_page3, json=resp_data_3, status=200, match_querystring=True)
    with patch('datarobot.models.reason_codes.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        reason_codes_rows = list(ReasonCodes(
            one_rcp.id, project_id, '', '', '', 0, 0, ''
        ).get_rows())
    assert len(reason_codes_rows) == 5
    for i, row in enumerate(reason_codes_rows):
        assert isinstance(row, ReasonCodesRow)
        assert row.row_id == one_rcp.data[i]['row_id']
        assert row.prediction == one_rcp.data[i]['prediction']
        assert row.prediction_values == one_rcp.data[i]['prediction_values']
        assert row.reason_codes == one_rcp.data[i]['reason_codes']


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_as_dataframe_regression(project_url, rcp_data, one_rc):
    url = '{}reasonCodes/{}/'.format(project_url, one_rc.id)
    responses.add(responses.GET, url, json=rcp_data, status=200)
    with patch('datarobot.models.reason_codes.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        reason_codes_as_dataframe = one_rc.get_all_as_dataframe()
    assert isinstance(reason_codes_as_dataframe, pandas.DataFrame)
    rows = rcp_data['count']
    # 5 columns per reason code, plus prediction and row id
    columns = 5 * one_rc.max_codes + 2
    assert reason_codes_as_dataframe.shape == (rows, columns)


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_as_dataframe_regression_with_exposure(project_url, one_rc,
                                                            rcp_data_with_adjusted_predictions):
    url = '{}reasonCodes/{}/'.format(project_url, one_rc.id)
    responses.add(responses.GET, url, json=rcp_data_with_adjusted_predictions, status=200)
    with patch('datarobot.models.reason_codes.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        reason_codes_as_dataframe = one_rc.get_all_as_dataframe()
    assert isinstance(reason_codes_as_dataframe, pandas.DataFrame)
    rows = rcp_data_with_adjusted_predictions['count']
    # 5 columns per reason code, plus prediction, adjusted prediction and row id
    columns = 5 * one_rc.max_codes + 3
    assert reason_codes_as_dataframe.shape == (rows, columns)


@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_reason_codes_as_dataframe_classification(project_url, rcp_data_classification, one_rc):
    url = '{}reasonCodes/{}/'.format(project_url, one_rc.id)
    responses.add(responses.GET, url, json=rcp_data_classification, status=200)
    with patch('datarobot.models.reason_codes.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        reason_codes_as_dataframe = one_rc.get_all_as_dataframe()
    assert isinstance(reason_codes_as_dataframe, pandas.DataFrame)
    rows = rcp_data_classification['count']
    # 5 columns per reason code, plus 2 columns per target level, plus row_id and prediction
    columns = (5 * one_rc.max_codes) + (2 * 2) + 2
    assert reason_codes_as_dataframe.shape == (rows, columns)


def test_reason_codes_download_to_csv():
    with patch('datarobot.models.reason_codes.ReasonCodes.get_all_as_dataframe') as rcmock:
        reason_codes = ReasonCodes('fake_id', 'fake_project_id', '', '', 3, 0, 0, '')
        reason_codes.download_to_csv('test_file.csv', encoding='ascii')
        rcmock.return_value.to_csv.assert_called_once_with(path_or_buf='test_file.csv',
                                                           header=True,
                                                           index=False,
                                                           encoding='ascii')



import json
import pandas as pd
import pytest
import responses
from datarobot import Predictions


@pytest.fixture
def sample_prediction():
    return {
        u'positiveClass': None,
        u'task': u'Regression',
        u'predictions': [
            {u'positiveProbability': None, u'prediction': 32.0,
             u'rowId': 0},
            {u'positiveProbability': None, u'prediction': 100.0,
             u'rowId': 1},
            {u'positiveProbability': None, u'prediction': 212.0,
             u'rowId': 2}
        ]
    }


@pytest.fixture
def prediction_id():
    return '59d92fe8962d7464ca8c6bd6'


@pytest.fixture
def model_id():
    return "modelid"


@pytest.fixture
def dataset_id():
    return "datasetid"


@responses.activate
def test_get__no_requests_made(project_id, prediction_id):
    Predictions.get(project_id, prediction_id)
    assert len(responses.calls) == 0


@responses.activate
@pytest.mark.usefixtures('client')
def test_list__ok(project_id, model_id, prediction_id, dataset_id):
    url = 'http://localhost/api/v2/projects/{}/predictions/{}/'.format(project_id,
                                                                       prediction_id)
    prediction_stub = {
        'url': 'https://host_name.com/projects/{}/predictions/{}//'.format(
            project_id,
            prediction_id,
        ),
        'id': prediction_id,
        'modelId': model_id,
        'datasetId': dataset_id,
        'url': url
    }
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/predictions/'.format(
            project_id,
        ),
        status=200,
        body=json.dumps({'data': [prediction_stub]}),
    )

    predictions_list = Predictions.list(project_id)

    assert len(predictions_list) == 1
    prediction = predictions_list[0]
    assert prediction.project_id == project_id
    assert prediction.model_id == model_id
    assert prediction.dataset_id == dataset_id
    assert prediction.prediction_id == prediction_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_all_as_dataframe__ok(
        project_id, prediction_id,
        project_url, project_with_target_json, sample_prediction
):
    url = 'https://host_name.com/projects/{}/predictions/{}/'.format(
        project_id,
        prediction_id,
    )
    responses.add(
        responses.GET,
        url,
        body=json.dumps(sample_prediction),
        status=200,
        content_type='application/json',
    )
    responses.add(
        responses.GET,
        project_url,
        body=project_with_target_json,
        status=200,
        content_type='application/json'
    )

    obj = Predictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe()

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (len(sample_prediction['predictions']), 2)
    assert list(data_frame.columns) == [
        'prediction',
        'row_id',
    ]


def test_repr():
    preds = Predictions('project_id', 'prediction_id', 'model_id', 'dataset_id')

    assert repr(preds) == "Predictions(prediction_id='prediction_id', project_id='project_id', " \
                          "model_id='model_id', dataset_id='dataset_id')"



import json

import pytest
import responses


@responses.activate
@pytest.mark.usefixtures('client')
def test_set_target(datetime_partition_spec, holdout_start_date,
                    project_with_target_json, project_without_target,
                    unittest_endpoint, project_id):
    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in datetime_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': datetime_partition_spec.datetime_partition_column,
        'useTimeSeries': datetime_partition_spec.use_time_series,
        'defaultToKnownInAdvance': datetime_partition_spec.default_to_known_in_advance,
        'autopilotDataSelectionMethod': datetime_partition_spec.autopilot_data_selection_method,
        'validationDuration': datetime_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': datetime_partition_spec.holdout_duration,
        'gapDuration': datetime_partition_spec.gap_duration,
        'numberOfBacktests': datetime_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto'
    }
    project_without_target.set_target('target', partitioning_method=datetime_partition_spec)

    assert responses.calls[0].request.method == 'PATCH'
    actual_request = json.loads(responses.calls[0].request.body)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
@pytest.mark.usefixtures('client')
def test_time_series_set_target(time_series_partition_spec, holdout_start_date,
                                project_with_target_json, project_without_target,
                                unittest_endpoint, project_id):
    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto'
    }
    project_without_target.set_target('target', partitioning_method=time_series_partition_spec)

    assert responses.calls[0].request.method == 'PATCH'
    actual_request = json.loads(responses.calls[0].request.body)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
@pytest.mark.usefixtures('client')
def test_multiseries_set_target(time_series_partition_spec, holdout_start_date,
                                project_with_target_json, project_without_target,
                                unittest_endpoint, project_id):
    time_series_partition_spec.multiseries_id_columns = ['series_id']
    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)
    multi_prop_url = '{}features/{}/multiseriesProperties/'.format(
        project_url, time_series_partition_spec.datetime_partition_column)
    multi_prop_json = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'detectedMultiseriesIdColumns': [
            {'multiseriesIdColumns': ['series_id'], 'timeStep': 1, 'timeUnit': 'DAY'}
        ]
    }
    responses.add(responses.GET, multi_prop_url,
                  json=multi_prop_json)
    date_feat_json = {
        'projectId': project_id, 'name': time_series_partition_spec.datetime_partition_column,
        'id': 2, 'timeSeriesEligible': False, 'featureType': 'date',
        'timeSeriesEligibilityReason': 'notUnique', 'lowInformation': False, 'uniqueCount': 500
    }
    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.datetime_partition_column),
                  json=date_feat_json)

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'multiseriesIdColumns': time_series_partition_spec.multiseries_id_columns,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto'
    }
    project_without_target.set_target('target', partitioning_method=time_series_partition_spec)

    assert responses.calls[2].request.method == 'PATCH'
    actual_request = json.loads(responses.calls[2].request.body)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload



import pytest

from datarobot.client import get_client


@pytest.mark.usefixtures('no_version_check')
def test_verify_parameter_false(default_config_endpoint,
                                default_config_token,
                                clean_user_config):
    """
    Can disable SSL verification. Not recommended save in rare cases.

    Parameters
    ----------
    default_config_endpoint
    default_config_token
    clean_user_config
    """
    configs = (
        'endpoint: {}'.format(default_config_endpoint),
        'token: {}'.format(default_config_token),
        'ssl_verify: false'
    )
    config = '\n'.join(configs)
    clean_user_config.write(config)
    c = get_client()
    assert not c.verify


@pytest.mark.usefixtures('no_version_check')
def test_verify_parameter_certfile(default_config_endpoint,
                                   default_config_token,
                                   clean_user_config):
    """
    Can specify a .pem CA file to use to verify server certificates

    Parameters
    ----------
    default_config_endpoint
    default_config_token
    clean_user_config
    """
    configs = (
        'endpoint: {}'.format(default_config_endpoint),
        'token: {}'.format(default_config_token),
        'ssl_verify: /path/to/certfile.pem'
    )
    config = '\n'.join(configs)
    clean_user_config.write(config)
    c = get_client()
    assert c.verify == '/path/to/certfile.pem'


@pytest.mark.usefixtures('default_config', 'no_version_check')
def test_ssl_verify_default():
    """
    By default, verify remains True (uses system information to verify
    certificate authenticity)
    """
    c = get_client()
    assert c.verify is True



import warnings
import trafaret as t

from datarobot.utils import from_api
from .api_object import APIObject
from ..utils import recognize_sourcedata, get_id_from_response, encode_utf8_if_py2
from ..errors import InvalidRatingTableWarning


class RatingTable(APIObject):
    """ Interface to modify and download rating tables.

    Attributes
    ----------
    id : str
        The id of the rating table.
    project_id : str
        The id of the project this rating table belongs to.
    rating_table_name : str
        The name of the rating table.
    original_filename : str
        The name of the file used to create the rating table.
    parent_model_id : str
        The model id of the model the rating table was validated against.
    model_id : str
        The model id of the model that was created from the rating table.
        Can be None if a model has not been created from the rating table.
    model_job_id : str
        The id of the job to create a model from this rating table.
        Can be None if a model has not been created from the rating table.
    validation_job_id : str
        The id of the created job to validate the rating table.
        Can be None if the rating table has not been validated.
    validation_error : str
        Contains a description of any errors caused during validation.
    """
    _converter = t.Dict({
        t.Key('id'): t.String,
        t.Key('project_id'): t.String,
        t.Key('rating_table_name', optional=True): t.String,
        t.Key('original_filename', optional=True): t.String,
        t.Key('parent_model_id', optional=True): t.String,
        t.Key('model_id', optional=True): t.String,
        t.Key('model_job_id', optional=True): t.String,
        t.Key('validation_job_id', optional=True): t.String,
        t.Key('validation_error', optional=True): t.String(allow_blank=True)
    }).allow_extra('*')

    def __init__(self, id, rating_table_name, original_filename, project_id, parent_model_id,
                 model_id=None, model_job_id=None, validation_job_id=None, validation_error=None):
        self.id = id
        self.rating_table_name = rating_table_name
        self.original_filename = original_filename
        self.project_id = project_id
        self.parent_model_id = parent_model_id
        self.model_id = model_id
        self.model_job_id = model_job_id
        self.validation_job_id = validation_job_id
        self.validation_error = validation_error

    @classmethod
    def from_server_data(cls, data, should_warn=True, keep_attrs=None):
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        should_warn : bool
            Whether or not to issue a warning if an invalid rating table is being retrieved.
        """
        case_converted = from_api(data, keep_attrs=keep_attrs)
        if should_warn:
            cls._warn_on_validation_error(case_converted['validation_error'])
        return cls.from_data(case_converted)

    def __repr__(self):
        return encode_utf8_if_py2(u'{}({})'.format(type(self).__name__, self.rating_table_name))

    @staticmethod
    def _warn_on_validation_error(validation_error):
        if validation_error != '':
            warnings.warn(u'The retrieved rating table was invalid, '
                          u'validation error: {}'.format(validation_error),
                          InvalidRatingTableWarning,
                          stacklevel=6)

    @classmethod
    def get(cls, project_id, rating_table_id):
        """Retrieve a single rating table

        Parameters
        ----------
        project_id : str
            The ID of the project the rating table is associated with.
        rating_table_id : str
            The ID of the rating table

        Returns
        -------
        rating_table : RatingTable
            The queried instance
        """
        path = 'projects/{}/ratingTables/{}/'.format(project_id, rating_table_id)
        rating_table = cls.from_location(path)
        return rating_table

    @classmethod
    def create(cls, project_id, parent_model_id, filename,
               rating_table_name='Uploaded Rating Table'):
        """
        Uploads and validates a new rating table CSV

        Parameters
        ----------
        project_id : str
            id of the project the rating table belongs to
        parent_model_id : str
            id of the model for which this rating table should be validated against
        filename : str
            The path of the CSV file containing the modified rating table.
        rating_table_name : str, optional
            A human friendly name for the new rating table. The string may be
            truncated and a suffix may be added to maintain unique names of all
            rating tables.

        Returns
        -------
        job: Job
            an instance of created async job

        Raises
        ------
        InputNotUnderstoodError
            Raised if `filename` isn't one of supported types.
        ClientError (400)
            Raised if `parent_model_id` is invalid.
        """
        from .job import Job

        form_data = {
            'parent_model_id': parent_model_id,
            'rating_table_name': rating_table_name,
        }
        path = 'projects/{}/ratingTables/'.format(project_id)

        default_fname = 'rating_table.csv'
        filesource_kwargs = recognize_sourcedata(filename, default_fname)
        response = cls._client.build_request_with_file(
            url=path,
            form_data=form_data,
            method='post',
            file_field_name='ratingTableFile',
            **filesource_kwargs)
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    def download(self, filepath):
        """
        Download a csv file containing the contents of this rating table

        Parameters
        ----------
        filepath : str
            The path at which to save the rating table file.
        """
        path = 'projects/{}/ratingTables/{}/file'.format(self.project_id, self.id)
        response = self._client.get(path)
        with open(filepath, mode='wb') as out_file:
            out_file.write(response.content)

    def rename(self, rating_table_name):
        """
        Renames a rating table to a different name.

        Parameters
        ----------
        rating_table_name : str
            The new name to rename the rating table to.
        """
        path = 'projects/{}/ratingTables/{}/'.format(self.project_id, self.id)
        response = self._client.patch(path, data={'ratingTableName': rating_table_name})
        updated_rating_table = self.from_server_data(response.json())
        self.rating_table_name = updated_rating_table.rating_table_name

    def create_model(self):
        """
        Creates a new model from this rating table record. This rating table
        must not already be associated with a model and must be valid.

        Returns
        -------
        job: Job
            an instance of created async job

        Raises
        ------
        ClientError (422)
            Raised if creating model from a RatingTable that failed validation
        JobAlreadyRequested
            Raised if creating model from a RatingTable that is already
            associated with a RatingTableModel
        """
        from .model import RatingTableModel
        return RatingTableModel.create_from_rating_table(self.project_id, self.id)



# -*- coding: utf-8 -*-
"""
    ext
    ~~~~~~~~~
    Borrowed very liberally from flask.ext

    :copyright: (c) 2015 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""


def setup():
    from ..exthook import ExtensionImporter
    importer = ExtensionImporter(['datarobot_%s'], __name__)
    importer.install()


setup()
del setup



import json

import pytest
import responses
from datarobot.errors import ClientError

from datarobot import Blueprint
from datarobot.models import (BlueprintChart,
                              BlueprintTaskDocument,
                              ModelBlueprintChart,)
from datarobot.utils import from_api


@pytest.fixture
def blueprint_id():
    return 'e1c7fc29ba2e612a72272324b8a842af'


@pytest.fixture
def blueprint_data(blueprint_id, project_id):
    return {'modelType': 'RandomForest Regressor',
            'processes': ['Missing values imputed'],
            'projectId': project_id,
            'id': blueprint_id,
            'blueprintCategory': 'DataRobot'
            }


@pytest.fixture
def blueprint_url(project_id, blueprint_id):
    return 'https://host_name.com/projects/{}/blueprints/{}/'.format(
        project_id,
        blueprint_id
    )


@pytest.fixture
def blueprint_chart_dot():
    return """digraph "Blueprint Chart" {
graph [rankdir=LR]
1 [label="Task1"]
2 [label="Task2"]
1 -> 2
}"""


@pytest.fixture
def blueprint_docs_url(project_id, blueprint_id):
    return 'https://host_name.com/projects/{}/blueprints/{}/blueprintDocs/'.format(
        project_id,
        blueprint_id
    )


@pytest.fixture()
def broken_blueprint_docs(blueprint_docs):
    bad_data = {'some_broken_field': 'some_broken_value'}
    bad_doc = blueprint_docs[0].copy()
    bad_doc.update(bad_data)
    bad_doc['parameters'][0].update(bad_data)
    bad_doc['references'][0].update(bad_data)
    bad_doc['links'][0].update(bad_data)
    return [bad_doc]


@pytest.fixture
def blueprint_chart_url(project_id, blueprint_id):
    return 'https://host_name.com/projects/{}/blueprints/{}/blueprintChart/'.format(
        project_id,
        blueprint_id
    )


def test_instantiation(blueprint_data):
    bp = Blueprint.from_data(from_api(blueprint_data))

    assert bp.model_type == blueprint_data['modelType']
    assert bp.processes == blueprint_data['processes']
    assert bp.project_id == blueprint_data['projectId']
    assert bp.id == blueprint_data['id']
    assert bp.blueprint_category == blueprint_data['blueprintCategory']


def test_future_proof(blueprint_data):
    Blueprint.from_data(dict(from_api(blueprint_data), new_key='future'))


def test_non_ascii(blueprint_data, unicode_string):
    bp = Blueprint.from_server_data(dict(blueprint_data, modelType=unicode_string))
    print(bp)  # test that __repr__ works correctly - this used to fail


@responses.activate
@pytest.mark.usefixtures('client')
def test_blueprint_get(blueprint_data, project_id, blueprint_url, blueprint_id):
    responses.add(
        responses.GET,
        blueprint_url,
        status=200,
        content_type='application/json',
        body=json.dumps(blueprint_data)
    )
    bp = Blueprint.get(project_id, blueprint_id)

    assert bp.model_type == blueprint_data['modelType']
    assert bp.processes == blueprint_data['processes']
    assert bp.project_id == blueprint_data['projectId']
    assert bp.id == blueprint_data['id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_blueprint_get_not_found(project_id, blueprint_url, blueprint_id):
    responses.add(
        responses.GET,
        blueprint_url,
        status=404,
        content_type='application/json',
        body=json.dumps({'message': 'No blueprint data found'})
    )
    with pytest.raises(ClientError):
        Blueprint.get(project_id, blueprint_id)


@responses.activate
@pytest.mark.usefixtures('client')
def test_blueprint_chart_get(blueprint_chart_data, project_id,
                             blueprint_chart_url, blueprint_id):
    responses.add(
        responses.GET,
        blueprint_chart_url,
        status=200,
        content_type='application/json',
        body=json.dumps(blueprint_chart_data)
    )
    bp = BlueprintChart.get(project_id, blueprint_id)

    assert bp.nodes == blueprint_chart_data['nodes']
    assert bp.edges == blueprint_chart_data['edges']


def test_blueprint_chart_to_graphviz(blueprint_chart_data, blueprint_chart_dot):
    bp = BlueprintChart(**blueprint_chart_data)
    assert bp.to_graphviz() == blueprint_chart_dot


@responses.activate
@pytest.mark.usefixtures('client')
def test_blueprint_get_documents(blueprint_docs_url, blueprint_docs,
                                 project_id, blueprint_id):
    responses.add(
        responses.GET,
        blueprint_docs_url,
        status=200,
        content_type='application/json',
        body=json.dumps(blueprint_docs)
    )
    bp = Blueprint.from_data({
        'id': blueprint_id,
        'project_id': project_id
    })
    docs = bp.get_documents()
    assert len(docs) == 1
    assert isinstance(docs[0], BlueprintTaskDocument)
    assert docs[0].title == blueprint_docs[0]['title']
    assert docs[0].task == blueprint_docs[0]['task']
    assert docs[0].description == blueprint_docs[0]['description']
    assert docs[0].links == blueprint_docs[0]['links']
    assert docs[0].parameters == blueprint_docs[0]['parameters']
    assert docs[0].references == blueprint_docs[0]['references']


@responses.activate
@pytest.mark.usefixtures('client')
def test_blueprint_get_documents_future_proof(blueprint_docs_url, broken_blueprint_docs,
                                              project_id, blueprint_id):
    responses.add(
        responses.GET,
        blueprint_docs_url,
        status=200,
        content_type='application/json',
        body=json.dumps(broken_blueprint_docs)
    )
    bp = Blueprint.from_data({
        'id': blueprint_id,
        'project_id': project_id
    })
    docs = bp.get_documents()
    assert len(docs) == 1
    assert isinstance(docs[0], BlueprintTaskDocument)


@responses.activate
@pytest.mark.usefixtures('client')
def test_model_blueprint_chart_get(blueprint_chart_data,
                                   project_id,
                                   model_id):
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/models/{}/blueprintChart/'.format(
            project_id,
            model_id
        ),
        status=200,
        content_type='application/json',
        body=json.dumps(blueprint_chart_data)
    )
    model_bp = ModelBlueprintChart.get(project_id, model_id)

    assert model_bp.nodes == blueprint_chart_data['nodes']
    assert model_bp.edges == blueprint_chart_data['edges']



def enum(*vals, **enums):
    """
    Enum without third party libs and compatible with py2 and py3 versions.
    """
    enums.update(dict(zip(vals, vals)))
    return type('Enum', (), enums)


CV_METHOD = enum(
    DATETIME='datetime',
    GROUP='group',
    RANDOM='random',
    STRATIFIED='stratified',
    USER='user',
)


SCORING_TYPE = enum(
    cross_validation='crossValidation',
    validation='validation',
)


DATETIME_AUTOPILOT_DATA_SELECTION_METHOD = enum(
    DURATION='duration',
    ROW_COUNT='rowCount',
)


VERBOSITY_LEVEL = enum(
    SILENT=0,
    VERBOSE=2,
)


# This is deprecated, to be removed in 3.0.
MODEL_JOB_STATUS = enum(
    ERROR='error',
    INPROGRESS='inprogress',
    QUEUE='queue',
)


# This is the job/queue status enum we want to keep.
# In 3.0 this will be INITIALIZED, RUNNING, ABORTED, COMPLETED, ERROR.
# And maybe the name will change to JobStatus.
QUEUE_STATUS = enum(
    ABORTED='ABORTED',
    COMPLETED='COMPLETED',
    ERROR='error',
    INPROGRESS='inprogress',
    QUEUE='queue',
)


AUTOPILOT_MODE = enum(
    FULL_AUTO='auto',
    MANUAL='manual',
    QUICK='quick',
)


PROJECT_STAGE = enum(
    AIM='aim',
    EDA='eda',
    EMPTY='empty',
    MODELING='modeling',
)


ASYNC_PROCESS_STATUS = enum(
    ABORTED='ABORTED',
    COMPLETED='COMPLETED',
    ERROR='ERROR',
    INITIALIZED='INITIALIZED',
    RUNNING='RUNNING',
)


LEADERBOARD_SORT_KEY = enum(
    PROJECT_METRIC='metric',
    SAMPLE_PCT='samplePct',
)


TARGET_TYPE = enum(
    BINARY='Binary',
    MULTICLASS='Multiclass',
    REGRESSION='Regression',
)


JOB_TYPE = enum(
    FEATURE_IMPACT='featureImpact',
    MODEL='model',
    MODEL_EXPORT='modelExport',
    PREDICT='predict',
    TRAINING_PREDICTIONS='trainingPredictions',
    PRIME_MODEL='primeModel',
    PRIME_RULESETS='primeRulesets',
    PRIME_VALIDATION='primeDownloadValidation',
    REASON_CODES='reasonCodes',
    REASON_CODES_INITIALIZATION='reasonCodesInitialization',
    PREDICTION_EXPLANATIONS='predictionExplanations',
    PREDICTION_EXPLANATIONS_INITIALIZATION='predictionExplanationsInitialization',
    RATING_TABLE_VALIDATION='validateRatingTable',
)


PREDICT_JOB_STATUS = enum(
    ABORTED='ABORTED',
    ERROR='error',
    INPROGRESS='inprogress',
    QUEUE='queue',
)

PREDICTION_PREFIX = enum(
    DEFAULT='class_'
)

PRIME_LANGUAGE = enum(
    JAVA='Java',
    PYTHON='Python',
)


VARIABLE_TYPE_TRANSFORM = enum(
    CATEGORICAL='categorical',
    CATEGORICAL_INT='categoricalInt',
    NUMERIC='numeric',
    TEXT='text',
)


DATE_EXTRACTION = enum(
    MONTH='month',
    MONTH_DAY='monthDay',
    WEEK='week',
    WEEK_DAY='weekDay',
    YEAR='year',
    YEAR_DAY='yearDay',
)


POSTGRESQL_DRIVER = enum(
    ANSI='PostgreSQL ANSI',
    UNICODE='PostgreSQL Unicode',
)

BLENDER_METHOD = enum(
    AVERAGE='AVG',
    ENET='ENET',
    GLM='GLM',
    MAE='MAE',
    MAEL1='MAEL1',
    MEDIAN='MED',
    PLS='PLS',
    RANDOM_FOREST='RF',
    LIGHT_GBM='LGBM',
    TENSORFLOW='TF'
)


CHART_DATA_SOURCE = enum(
    CROSSVALIDATION='crossValidation',
    HOLDOUT='holdout',
    VALIDATION='validation',
)


SCALEOUT_MODELING_MODE = enum(
    AUTOPILOT='autopilot',
    DISABLED='disabled',
    REPOSITORY_ONLY='repositoryOnly',
)


DATA_SUBSET = enum(
    ALL='all',
    VALIDATION_AND_HOLDOUT='validationAndHoldout',
    HOLDOUT='holdout',
)

DEFAULT_MAX_WAIT = 600

# default time out values in seconds for waiting response from client
DEFAULT_TIMEOUT = enum(
    CONNECT=6.05,  # time in seconds for the connection to server to be established
    READ=60,  # time in seconds after which to conclude the server isn't responding anymore
    UPLOAD=600,  # time in seconds after which to conclude that project dataset cannot be uploaded
)

# Time in seconds after which to conclude the server isn't responding anymore
# same as in DEFAULT_TIMEOUT, keeping for backwards compatibility
DEFAULT_READ_TIMEOUT = DEFAULT_TIMEOUT.READ

TARGET_LEAKAGE_TYPE = enum(
    SKIPPED_DETECTION='SKIPPED_DETECTION',
    FALSE='FALSE',
    MODERATE_RISK='MODERATE_RISK',
    HIGH_RISK='HIGH_RISK'
)

TREAT_AS_EXPONENTIAL = enum(
    ALWAYS='always',
    NEVER='never',
    AUTO='auto'
)

DIFFERENCING_METHOD = enum(
    AUTO='auto',
    SIMPLE='simple',
    NONE='none',
    SEASONAL='seasonal'
)

PERIODICITY_TIME_UNITS = enum(
    SECOND='SECOND',
    MINUTE='MINUTE',
    HOUR='HOUR',
    DAY='DAY',
    WEEK='WEEK',
    MONTH='MONTH',
    QUARTER='QUARTER',
    YEAR='YEAR')

PERIODICITY_MAX_TIME_STEP = 9223372036854775807

RECOMMENDED_MODEL_TYPE = enum(
    MOST_ACCURATE='Most Accurate',
    LIMITED_ACCURATE='Recommended',
    FAST_ACCURATE='Fast & Accurate',
    RECOMMENDED_FOR_DEPLOYMENT='Recommended for Deployment'
)

MONOTONICITY_FEATURELIST_DEFAULT = object()



import pytest
import responses
import json

from datarobot import ModelRecommendation
from datarobot.enums import RECOMMENDED_MODEL_TYPE


@pytest.fixture
def list_recommended_model_json():
    model_list = [{
        'recommendationType': RECOMMENDED_MODEL_TYPE.MOST_ACCURATE,
        'modelId': 'model1',
        'projectId': 'project1'
    }, {
        'recommendationType': RECOMMENDED_MODEL_TYPE.LIMITED_ACCURATE,
        'modelId': 'model2',
        'projectId': 'project1'
    }]
    return json.dumps(model_list)


@pytest.fixture
def recommended_model_json():
    model = {
        'recommendationType': 'Recommended',
        'modelId': '5ad8e121fcf4c87d96ae02dd',
        'projectId': 'project1'
    }
    return json.dumps(model)


@pytest.fixture
def recommended_model_extra_field_json():
    model = {
        'recommendationType': 'Recommended',
        'modelId': '5ad8e121fcf4c87d96ae02dd',
        'projectId': 'project1',
        'extraField': 'value'
    }
    return json.dumps(model)


@responses.activate
@pytest.mark.usefixtures('client')
def test_recommended_model_default(recommended_model_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/recommendedModels/recommendedModel/',
                  body=recommended_model_json)
    recommended_model = ModelRecommendation.get('p-id')
    assert recommended_model.recommendation_type == 'Recommended'
    assert recommended_model.model_id == '5ad8e121fcf4c87d96ae02dd'
    assert recommended_model.project_id == 'project1'


@responses.activate
@pytest.mark.usefixtures('client')
def test_recommended_model_specific(list_recommended_model_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/recommendedModels/',
                  body=list_recommended_model_json)
    recommended_model = ModelRecommendation.get('p-id', RECOMMENDED_MODEL_TYPE.MOST_ACCURATE)
    assert recommended_model.recommendation_type == 'Most Accurate'
    assert recommended_model.model_id == 'model1'
    assert recommended_model.project_id == 'project1'


@responses.activate
@pytest.mark.usefixtures('client')
def test_recommended_model_ignore_extra_fields(recommended_model_extra_field_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/recommendedModels/recommendedModel/',
                  body=recommended_model_extra_field_json)
    recommended_model = ModelRecommendation.get('p-id')
    assert recommended_model.recommendation_type == 'Recommended'
    assert recommended_model.model_id == '5ad8e121fcf4c87d96ae02dd'
    assert recommended_model.project_id == 'project1'
    assert not hasattr(recommended_model, 'extraField')


@responses.activate
@pytest.mark.usefixtures('client')
def test_recommended_model_list(list_recommended_model_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/recommendedModels/',
                  body=list_recommended_model_json)
    recommended_models = ModelRecommendation.get_all('p-id')
    assert len(recommended_models) == 2
    expected_models = json.loads(list_recommended_model_json)

    expected_model1 = next(model for model in expected_models if model['modelId'] == 'model1')
    expected_model2 = next(model for model in expected_models if model['modelId'] == 'model2')
    actual_model1 = [model for model in recommended_models if model.model_id == 'model1']
    actual_model2 = [model for model in recommended_models if model.model_id == 'model2']

    assert len(actual_model1) == 1
    assert len(actual_model2) == 1
    actual_model1 = actual_model1[0]
    actual_model2 = actual_model2[0]

    assert actual_model1.recommendation_type == expected_model1['recommendationType']
    assert actual_model2.recommendation_type == expected_model2['recommendationType']
    assert actual_model1.project_id == expected_model1['projectId']
    assert actual_model2.project_id == expected_model2['projectId']

    assert ModelRecommendation.get_recommendation(
        recommended_models,
        RECOMMENDED_MODEL_TYPE.MOST_ACCURATE) is not None

    assert ModelRecommendation.get_recommendation(
        recommended_models,
        RECOMMENDED_MODEL_TYPE.FAST_ACCURATE) is None
